#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import unittest

from mock import MagicMock, patch

from neptune.internal.hardware.constants import BYTES_IN_ONE_GB
from neptune.internal.hardware.gauges.gpu import GpuMemoryGauge, GpuUsageGauge


@patch('neptune.internal.hardware.gpu.gpu_monitor.nvmlInit', MagicMock())
class TestGPUGauges(unittest.TestCase):
    def setUp(self):
        self.card_index = 2
        self.gpu_card_handle = MagicMock()

        patcher = patch('neptune.internal.hardware.gpu.gpu_monitor.nvmlDeviceGetHandleByIndex')
        nvmlDeviceGetHandleByIndex = patcher.start()
        nvmlDeviceGetHandleByIndex.side_effect = \
            lambda card_index: self.gpu_card_handle if card_index == self.card_index else None
        self.addCleanup(patcher.stop)

    @patch('neptune.internal.hardware.gpu.gpu_monitor.nvmlDeviceGetUtilizationRates')
    def test_gpu_usage_gauge(self, nvmlDeviceGetMemoryInfo):
        # given
        gauge = GpuUsageGauge(card_index=self.card_index)
        # and
        util_info = MagicMock()
        util_info.gpu = 40
        nvmlDeviceGetMemoryInfo.side_effect = lambda handle: util_info if handle == self.gpu_card_handle else None

        # when
        usage_percent = gauge.value()

        # then
        self.assertEqual(40.0, usage_percent)
        self.assertEqual(float, type(usage_percent))

    @patch('neptune.internal.hardware.gpu.gpu_monitor.nvmlDeviceGetMemoryInfo')
    def test_gpu_memory_gauge(self, nvmlDeviceGetMemoryInfo):
        # given
        gauge = GpuMemoryGauge(card_index=self.card_index)
        # and
        memory_info = MagicMock()
        memory_info.used = 3 * BYTES_IN_ONE_GB
        nvmlDeviceGetMemoryInfo.side_effect = lambda handle: memory_info if handle == self.gpu_card_handle else None

        # when
        memory_gb = gauge.value()

        # then
        self.assertEqual(memory_gb, 3.0)
        self.assertEqual(float, type(memory_gb))


if __name__ == '__main__':
    unittest.main()



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import abc
from typing import TypeVar, Generic

from neptune.new.internal.operation import (
    AssignArtifact,
    AddStrings,
    AssignBool,
    AssignDatetime,
    AssignFloat,
    AssignInt,
    AssignString,
    ClearArtifact,
    ClearFloatLog,
    ClearImageLog,
    ClearStringLog,
    ClearStringSet,
    ConfigFloatSeries,
    DeleteAttribute,
    DeleteFiles,
    LogFloats,
    LogImages,
    LogStrings,
    Operation,
    RemoveStrings,
    TrackFilesToArtifact,
    UploadFile,
    UploadFileContent,
    UploadFileSet,
)

Ret = TypeVar('Ret')


class OperationVisitor(Generic[Ret]):

    def visit(self, op: Operation) -> Ret:
        return op.accept(self)

    @abc.abstractmethod
    def visit_assign_float(self, op: AssignFloat) -> Ret:
        pass

    @abc.abstractmethod
    def visit_assign_int(self, op: AssignInt) -> Ret:
        pass

    @abc.abstractmethod
    def visit_assign_bool(self, op: AssignBool) -> Ret:
        pass

    @abc.abstractmethod
    def visit_assign_string(self, op: AssignString) -> Ret:
        pass

    @abc.abstractmethod
    def visit_assign_datetime(self, op: AssignDatetime) -> Ret:
        pass

    @abc.abstractmethod
    def visit_assign_artifact(self, op: AssignArtifact) -> Ret:
        pass

    @abc.abstractmethod
    def visit_upload_file(self, op: UploadFile) -> Ret:
        pass

    @abc.abstractmethod
    def visit_upload_file_content(self, op: UploadFileContent) -> Ret:
        pass

    @abc.abstractmethod
    def visit_upload_file_set(self, op: UploadFileSet) -> Ret:
        pass

    @abc.abstractmethod
    def visit_log_floats(self, op: LogFloats) -> Ret:
        pass

    @abc.abstractmethod
    def visit_log_strings(self, op: LogStrings) -> Ret:
        pass

    @abc.abstractmethod
    def visit_log_images(self, op: LogImages) -> Ret:
        pass

    @abc.abstractmethod
    def visit_clear_float_log(self, op: ClearFloatLog) -> Ret:
        pass

    @abc.abstractmethod
    def visit_clear_string_log(self, op: ClearStringLog) -> Ret:
        pass

    @abc.abstractmethod
    def visit_clear_image_log(self, op: ClearImageLog) -> Ret:
        pass

    @abc.abstractmethod
    def visit_config_float_series(self, op: ConfigFloatSeries) -> Ret:
        pass

    @abc.abstractmethod
    def visit_add_strings(self, op: AddStrings) -> Ret:
        pass

    @abc.abstractmethod
    def visit_remove_strings(self, op: RemoveStrings) -> Ret:
        pass

    @abc.abstractmethod
    def visit_delete_attribute(self, op: DeleteAttribute) -> Ret:
        pass

    @abc.abstractmethod
    def visit_clear_string_set(self, op: ClearStringSet) -> Ret:
        pass

    @abc.abstractmethod
    def visit_delete_files(self, op: DeleteFiles) -> Ret:
        pass

    @abc.abstractmethod
    def visit_track_files_to_artifact(self, op: TrackFilesToArtifact) -> Ret:
        pass

    @abc.abstractmethod
    def visit_clear_artifact(self, op: ClearArtifact) -> Ret:
        pass



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import re
import os
from typing import Optional, List, Dict

from bravado.client import SwaggerClient
from bravado.exception import HTTPNotFound, HTTPBadRequest, HTTPConflict, HTTPForbidden, HTTPUnprocessableEntity

from neptune.patterns import PROJECT_QUALIFIED_NAME_PATTERN
from neptune.new.envs import API_TOKEN_ENV_NAME
from neptune.new.internal.utils import verify_type
from neptune.new.internal.credentials import Credentials
from neptune.new.internal.backends.hosted_client import (
    create_backend_client,
    create_http_client_with_auth,
    DEFAULT_REQUEST_KWARGS,
)
from neptune.new.internal.backends.utils import with_api_exceptions_handler, ssl_verify, parse_validation_errors
from neptune.management.internal.utils import normalize_project_name
from neptune.management.internal.types import *
from neptune.management.exceptions import (
    AccessRevokedOnDeletion,
    AccessRevokedOnMemberRemoval,
    ProjectAlreadyExists,
    ProjectNotFound,
    UserNotExistsOrWithoutAccess,
    WorkspaceNotFound,
    UserAlreadyHasAccess,
    BadRequestException,
    ProjectsLimitReached,
)


def _get_token(api_token: Optional[str] = None) -> str:
    return api_token or os.getenv(API_TOKEN_ENV_NAME)


def _get_backend_client(api_token: Optional[str] = None) -> SwaggerClient:
    credentials = Credentials.from_token(api_token=_get_token(api_token=api_token))
    http_client, client_config = create_http_client_with_auth(
        credentials=credentials,
        ssl_verify=ssl_verify(),
        proxies={}
    )
    return create_backend_client(
        client_config=client_config,
        http_client=http_client
    )


@with_api_exceptions_handler
def get_project_list(api_token: Optional[str] = None) -> List[str]:
    verify_type('api_token', api_token, (str, type(None)))

    backend_client = _get_backend_client(api_token=api_token)

    params = {
        'userRelation': 'viewerOrHigher',
        'sortBy': ['lastViewed'],
        **DEFAULT_REQUEST_KWARGS
    }

    projects = backend_client.api.listProjects(**params).response().result.entries

    return [normalize_project_name(name=project.name, workspace=project.organizationName) for project in projects]


@with_api_exceptions_handler
def create_project(
        name: str,
        key: str,
        workspace: Optional[str] = None,
        visibility: ProjectVisibility = ProjectVisibility.PRIVATE,
        description: Optional[str] = None,
        api_token: Optional[str] = None
) -> str:
    verify_type('name', name, str)
    verify_type('key', key, str)
    verify_type('workspace', workspace, (str, type(None)))
    verify_type('visibility', visibility, ProjectVisibility)
    verify_type('description', description, (str, type(None)))
    verify_type('api_token', api_token, (str, type(None)))

    backend_client = _get_backend_client(api_token=api_token)
    project_identifier = normalize_project_name(name=name, workspace=workspace)

    project_spec = re.search(PROJECT_QUALIFIED_NAME_PATTERN, project_identifier)
    workspace, name = project_spec['workspace'], project_spec['project']

    try:
        workspaces = backend_client.api.listOrganizations(**DEFAULT_REQUEST_KWARGS).response().result
        workspace_name_to_id = {f'{f.name}': f.id for f in workspaces}
    except HTTPNotFound:
        raise WorkspaceNotFound(workspace=workspace)

    if workspace not in workspace_name_to_id:
        raise WorkspaceNotFound(workspace=workspace)

    params = {
        'projectToCreate': {
            'name': name,
            'description': description,
            'projectKey': key,
            'organizationId': workspace_name_to_id[workspace],
            'visibility': visibility.value
        },
        **DEFAULT_REQUEST_KWARGS
    }

    try:
        response = backend_client.api.createProject(**params).response()
        return normalize_project_name(name=response.result.name, workspace=response.result.organizationName)
    except HTTPBadRequest as e:
        validation_errors = parse_validation_errors(error=e)
        if 'ERR_NOT_UNIQUE' in validation_errors:
            raise ProjectAlreadyExists(name=project_identifier) from e
        raise BadRequestException(validation_errors=validation_errors)
    except HTTPUnprocessableEntity as e:
        raise ProjectsLimitReached() from e


@with_api_exceptions_handler
def delete_project(name: str, workspace: Optional[str] = None, api_token: Optional[str] = None):
    verify_type('name', name, str)
    verify_type('workspace', workspace, (str, type(None)))
    verify_type('api_token', api_token, (str, type(None)))

    backend_client = _get_backend_client(api_token=api_token)
    project_identifier = normalize_project_name(name=name, workspace=workspace)

    params = {
        'projectIdentifier': project_identifier,
        **DEFAULT_REQUEST_KWARGS
    }

    try:
        backend_client.api.deleteProject(**params).response()
    except HTTPNotFound as e:
        raise ProjectNotFound(name=project_identifier) from e
    except HTTPForbidden as e:
        raise AccessRevokedOnDeletion(name=project_identifier) from e


@with_api_exceptions_handler
def add_project_member(
        name: str,
        username: str,
        role: MemberRole,
        workspace: Optional[str] = None,
        api_token: Optional[str] = None
):
    verify_type('name', name, str)
    verify_type('username', username, str)
    verify_type('role', role, MemberRole)
    verify_type('workspace', workspace, (str, type(None)))
    verify_type('api_token', api_token, (str, type(None)))

    backend_client = _get_backend_client(api_token=api_token)
    project_identifier = normalize_project_name(name=name, workspace=workspace)

    params = {
        'projectIdentifier': project_identifier,
        'member': {
            'userId': username,
            'role': role.value
        },
        **DEFAULT_REQUEST_KWARGS
    }

    try:
        backend_client.api.addProjectMember(**params).response()
    except HTTPNotFound as e:
        raise ProjectNotFound(name=project_identifier) from e
    except HTTPConflict as e:
        raise UserAlreadyHasAccess(user=username, project=project_identifier) from e


@with_api_exceptions_handler
def get_project_member_list(
        name: str,
        workspace: Optional[str] = None,
        api_token: Optional[str] = None
) -> Dict[str, str]:
    verify_type('name', name, str)
    verify_type('workspace', workspace, (str, type(None)))
    verify_type('api_token', api_token, (str, type(None)))

    backend_client = _get_backend_client(api_token=api_token)
    project_identifier = normalize_project_name(name=name, workspace=workspace)

    params = {
        'projectIdentifier': project_identifier,
        **DEFAULT_REQUEST_KWARGS
    }

    try:
        result = backend_client.api.listProjectMembers(**params).response().result
        return {f'{m.registeredMemberInfo.username}': m.role for m in result}
    except HTTPNotFound as e:
        raise ProjectNotFound(name=project_identifier) from e


@with_api_exceptions_handler
def remove_project_member(
        name: str,
        username: str,
        workspace: Optional[str] = None,
        api_token: Optional[str] = None
):
    verify_type('name', name, str)
    verify_type('username', username, str)
    verify_type('workspace', workspace, (str, type(None)))
    verify_type('api_token', api_token, (str, type(None)))

    backend_client = _get_backend_client(api_token=api_token)
    project_identifier = normalize_project_name(name=name, workspace=workspace)

    params = {
        'projectIdentifier': project_identifier,
        'userId': username,
        **DEFAULT_REQUEST_KWARGS
    }

    try:
        backend_client.api.deleteProjectMember(**params).response()
    except HTTPNotFound as e:
        raise ProjectNotFound(name=project_identifier) from e
    except HTTPUnprocessableEntity as e:
        raise UserNotExistsOrWithoutAccess(user=username, project=project_identifier) from e
    except HTTPForbidden as e:
        raise AccessRevokedOnMemberRemoval(user=username, project=project_identifier) from e


@with_api_exceptions_handler
def get_workspace_member_list(name: str, api_token: Optional[str] = None) -> Dict[str, str]:
    verify_type('name', name, str)
    verify_type('api_token', api_token, (str, type(None)))

    backend_client = _get_backend_client(api_token=api_token)

    params = {
        'organizationIdentifier': name,
        **DEFAULT_REQUEST_KWARGS
    }

    try:
        result = backend_client.api.listOrganizationMembers(**params).response().result
        return {f'{m.registeredMemberInfo.username}': m.role for m in result}
    except HTTPNotFound as e:
        raise WorkspaceNotFound(workspace=name) from e



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import datetime
import tempfile
import unittest
from pathlib import Path

import boto3
import freezegun
from moto import mock_s3

from neptune.new.exceptions import NeptuneUnsupportedArtifactFunctionalityException
from neptune.new.internal.artifacts.types import ArtifactDriversMap, ArtifactFileData, ArtifactFileType
from neptune.new.internal.artifacts.drivers.s3 import S3ArtifactDriver

from tests.neptune.new.internal.artifacts.utils import md5


@mock_s3
class TestS3ArtifactDrivers(unittest.TestCase):
    def setUp(self):
        self.bucket_name = 'kuiper_belt'
        self.s3 = boto3.client('s3')
        self.s3.create_bucket(Bucket=self.bucket_name)
        self.update_time = datetime.datetime(2021, 5, 23, 3, 55, 26)
        with freezegun.freeze_time(self.update_time):
            self.s3.put_object(Bucket=self.bucket_name, Key='path/to/file1', Body=b'\xde\xad\xbe\xef')
            self.s3.put_object(Bucket=self.bucket_name, Key='path/to/file2', Body=b'\x20')
            self.s3.put_object(Bucket=self.bucket_name, Key='path/file3', Body=b'\x21')

    def test_match_by_path(self):
        self.assertEqual(
            ArtifactDriversMap.match_path(f"s3://{self.bucket_name}/path/to/"),
            S3ArtifactDriver
        )

    def test_match_by_type(self):
        self.assertEqual(
            ArtifactDriversMap.match_type("S3"),
            S3ArtifactDriver
        )

    def test_file_download(self):
        artifact_file = ArtifactFileData(
            file_path='to/file1',
            file_hash='2f249230a8e7c2bf6005ccd2679259ec',
            type=ArtifactFileType.S3.value,
            metadata={
                'location': f"s3://{self.bucket_name}/path/to/file1"
            }
        )

        with tempfile.TemporaryDirectory() as temporary:
            local_destination = Path(temporary) / "target.txt"

            S3ArtifactDriver.download_file(
                destination=local_destination,
                file_definition=artifact_file
            )

            self.assertEqual('2f249230a8e7c2bf6005ccd2679259ec', md5(local_destination))

    def test_single_retrieval(self):
        files = S3ArtifactDriver.get_tracked_files(f"s3://{self.bucket_name}/path/to/file1")

        self.assertEqual(1, len(files))
        self.assertIsInstance(files[0], ArtifactFileData)
        self.assertEqual(ArtifactFileType.S3.value, files[0].type)
        self.assertEqual('2f249230a8e7c2bf6005ccd2679259ec', files[0].file_hash)
        self.assertEqual('file1', files[0].file_path)
        self.assertEqual(4, files[0].size)
        self.assertEqual(
            {'location', 'last_modified'},
            files[0].metadata.keys()
        )
        self.assertEqual(
            f's3://{self.bucket_name}/path/to/file1',
            files[0].metadata['location']
        )
        self.assertEqual(
            self.update_time.strftime(S3ArtifactDriver.DATETIME_FORMAT),
            files[0].metadata['last_modified']
        )

    def test_multiple_retrieval(self):
        files = S3ArtifactDriver.get_tracked_files(f"s3://{self.bucket_name}/path/to/")
        files = sorted(files, key=lambda file: file.file_path)

        self.assertEqual(2, len(files))

        self.assertEqual('2f249230a8e7c2bf6005ccd2679259ec', files[0].file_hash)
        self.assertEqual('file1', files[0].file_path)
        self.assertEqual(f's3://{self.bucket_name}/path/to/file1', files[0].metadata['location'])

        self.assertEqual('7215ee9c7d9dc229d2921a40e899ec5f', files[1].file_hash)
        self.assertEqual('file2', files[1].file_path)
        self.assertEqual(f's3://{self.bucket_name}/path/to/file2', files[1].metadata['location'])

    def test_multiple_retrieval_prefix(self):
        files = S3ArtifactDriver.get_tracked_files(f"s3://{self.bucket_name}/path/", 'my/custom_path')
        files = sorted(files, key=lambda file: file.file_path)

        self.assertEqual(len(files), 3)

        self.assertEqual('9033e0e305f247c0c3c80d0c7848c8b3', files[0].file_hash)
        self.assertEqual('my/custom_path/file3', files[0].file_path)
        self.assertEqual(f's3://{self.bucket_name}/path/file3', files[0].metadata['location'])

        self.assertEqual('2f249230a8e7c2bf6005ccd2679259ec', files[1].file_hash)
        self.assertEqual('my/custom_path/to/file1', files[1].file_path)
        self.assertEqual(f's3://{self.bucket_name}/path/to/file1', files[1].metadata['location'])

        self.assertEqual('7215ee9c7d9dc229d2921a40e899ec5f', files[2].file_hash)
        self.assertEqual('my/custom_path/to/file2', files[2].file_path)
        self.assertEqual(f's3://{self.bucket_name}/path/to/file2', files[2].metadata['location'])

    def test_wildcards_not_supported(self):
        with self.assertRaises(NeptuneUnsupportedArtifactFunctionalityException):
            S3ArtifactDriver.get_tracked_files(f"s3://{self.bucket_name}/*/to/")



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import logging
import warnings
from typing import Optional

from neptune.new.types.atoms import GitRef

_logger = logging.getLogger(__name__)


def get_git_repo(repo_path):
    # WARN: GitPython asserts the existence of `git` executable
    # which consists in failure during the preparation of conda package
    try:
        # pylint:disable=import-outside-toplevel
        import git
        return git.Repo(repo_path, search_parent_directories=True)
    except ImportError:
        warnings.warn("GitPython could not be initialized")


def get_git_info(repo_path=None):
    try:
        # pylint:disable=bad-option-value

        repo = get_git_repo(repo_path)

        commit = repo.head.commit

        active_branch = ""

        try:
            active_branch = repo.active_branch.name
        except TypeError as e:
            if str(e.args[0]).startswith("HEAD is a detached symbolic reference as it points to"):
                active_branch = "Detached HEAD"

        remote_urls = [remote.url for remote in repo.remotes]

        return GitRef(
            commit_id=commit.hexsha,
            message=commit.message,
            author_name=commit.author.name,
            author_email=commit.author.email,
            commit_date=commit.committed_datetime,
            dirty=repo.is_dirty(untracked_files=True),
            branch=active_branch,
            remotes=remote_urls
        )
    except:  # pylint: disable=bare-except
        return None


def get_git_repo_path(initial_path: str) -> Optional[str]:
    try:
        return get_git_repo(initial_path).git_dir
    except:  # pylint: disable=bare-except
        pass


def discover_git_repo_location() -> Optional[str]:
    git_path = None

    # pylint:disable=bad-option-value,import-outside-toplevel
    import __main__

    if hasattr(__main__, '__file__'):
        git_path = get_git_repo_path(
            initial_path=os.path.dirname(os.path.abspath(__main__.__file__))
        )

    if not git_path:
        git_path = get_git_repo_path(initial_path=os.getcwd())

    return git_path



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import uuid
import unittest

from mock import patch, MagicMock, Mock
from bravado.exception import HTTPNotFound, HTTPForbidden, HTTPConflict, HTTPBadRequest, HTTPUnprocessableEntity
from bravado.testing.response_mocks import BravadoResponseMock

from neptune.management import (
    get_project_list,
    get_workspace_member_list,
    get_project_member_list,
    delete_project,
    create_project,
    add_project_member,
    remove_project_member,
    MemberRole,
)
from neptune.management.exceptions import (
    WorkspaceNotFound,
    ProjectNotFound,
    AccessRevokedOnDeletion,
    ProjectAlreadyExists,
    UserNotExistsOrWithoutAccess,
    UserAlreadyHasAccess,
    AccessRevokedOnMemberRemoval,
)
from neptune.new.internal.backends.utils import verify_host_resolution
from neptune.new.internal.backends.hosted_client import (
    _get_token_client,  # pylint:disable=protected-access
    get_client_config,
    create_http_client_with_auth,
    create_backend_client,
    create_leaderboard_client,
    create_artifacts_client,
)
from tests.neptune.new.backend_test_mixin import BackendTestMixin

API_TOKEN = 'eyJhcGlfYWRkcmVzcyI6Imh0dHBzOi8vYXBwLnN0YWdlLm5lcHR1bmUubWwiLCJ' \
            'hcGlfa2V5IjoiOTJhNzhiOWQtZTc3Ni00ODlhLWI5YzEtNzRkYmI1ZGVkMzAyIn0='


@patch('neptune.new.internal.backends.hosted_client.RequestsClient', new=MagicMock())
@patch('neptune.new.internal.backends.hosted_client.NeptuneAuthenticator', new=MagicMock())
@patch('bravado.client.SwaggerClient.from_url')
@patch('platform.platform', new=lambda: 'testPlatform')
@patch('platform.python_version', new=lambda: '3.9.test')
class TestHostedClient(unittest.TestCase, BackendTestMixin):

    def setUp(self) -> None:
        # Clear all LRU storage
        verify_host_resolution.cache_clear()
        _get_token_client.cache_clear()
        get_client_config.cache_clear()
        create_http_client_with_auth.cache_clear()
        create_backend_client.cache_clear()
        create_leaderboard_client.cache_clear()
        create_artifacts_client.cache_clear()

    def test_project_listing_empty(self, swagger_client_factory):
        swagger_client = self._get_swagger_client_mock(swagger_client_factory)

        # given:
        swagger_client.api.listProjects.return_value.response = BravadoResponseMock(
            result=Mock(
                entries=[]
            )
        )

        # when:
        returned_projects = get_project_list(api_token=API_TOKEN)

        # then:
        self.assertEqual([], returned_projects)

    def test_project_listing(self, swagger_client_factory):
        swagger_client = self._get_swagger_client_mock(swagger_client_factory)

        # given:
        project1 = MagicMock(
            organizationName='org1'
        )
        project1.name = 'project1'
        project2 = MagicMock(
            organizationName='org2'
        )
        project2.name = 'project2'
        projects = Mock(
            entries=[
                project1,
                project2
            ]
        )
        swagger_client.api.listProjects.return_value.response = BravadoResponseMock(
            result=projects,
        )

        # when:
        returned_projects = get_project_list(api_token=API_TOKEN)

        # then:
        self.assertEqual(['org1/project1', 'org2/project2'], returned_projects)

    def test_workspace_members(self, swagger_client_factory):
        swagger_client = self._get_swagger_client_mock(swagger_client_factory)

        # given:
        members = [
            Mock(
                role='member',
                registeredMemberInfo=Mock(
                    username='tester1'
                )
            ),
            Mock(
                role='admin',
                registeredMemberInfo=Mock(
                    username='tester2'
                )
            )
        ]
        swagger_client.api.listOrganizationMembers.return_value.response = BravadoResponseMock(
            result=members,
        )

        # when:
        returned_members = get_workspace_member_list(name='org2', api_token=API_TOKEN)

        # then:
        self.assertEqual({'tester1': 'member', 'tester2': 'admin'}, returned_members)

    def test_workspace_members_empty(self, swagger_client_factory):
        swagger_client = self._get_swagger_client_mock(swagger_client_factory)

        # given:
        members = []
        swagger_client.api.listOrganizationMembers.return_value.response = BravadoResponseMock(
            result=members,
        )

        # when:
        returned_members = get_workspace_member_list(name='org2', api_token=API_TOKEN)

        # then:
        self.assertEqual({}, returned_members)

    def test_workspace_members_not_exists(self, swagger_client_factory):
        swagger_client = self._get_swagger_client_mock(swagger_client_factory)

        # when:
        swagger_client.api.listOrganizationMembers.side_effect = HTTPNotFound(response=MagicMock())

        # then:
        with self.assertRaises(WorkspaceNotFound):
            get_workspace_member_list(name='org2', api_token=API_TOKEN)

    def test_project_members(self, swagger_client_factory):
        swagger_client = self._get_swagger_client_mock(swagger_client_factory)

        # given:
        members = [
            Mock(
                role='member',
                registeredMemberInfo=Mock(
                    username='tester1'
                )
            ),
            Mock(
                role='admin',
                registeredMemberInfo=Mock(
                    username='tester2'
                )
            )
        ]
        swagger_client.api.listProjectMembers.return_value.response = BravadoResponseMock(
            result=members,
        )

        # when:
        returned_members = get_project_member_list(name='org/proj', api_token=API_TOKEN)

        # then:
        self.assertEqual({'tester1': 'member', 'tester2': 'admin'}, returned_members)

    def test_project_members_empty(self, swagger_client_factory):
        swagger_client = self._get_swagger_client_mock(swagger_client_factory)

        # given:
        members = []
        swagger_client.api.listProjectMembers.return_value.response = BravadoResponseMock(
            result=members,
        )

        # when:
        returned_members = get_project_member_list(name='org/proj', api_token=API_TOKEN)

        # then:
        self.assertEqual({}, returned_members)

    def test_project_members_not_exists(self, swagger_client_factory):
        swagger_client = self._get_swagger_client_mock(swagger_client_factory)

        # when:
        swagger_client.api.listProjectMembers.side_effect = HTTPNotFound(response=MagicMock())

        # then:
        with self.assertRaises(ProjectNotFound):
            get_project_member_list(name='org/proj', api_token=API_TOKEN)

    def test_delete_project_not_exists(self, swagger_client_factory):
        swagger_client = self._get_swagger_client_mock(swagger_client_factory)

        # when:
        swagger_client.api.deleteProject.side_effect = HTTPNotFound(response=MagicMock())

        # then:
        with self.assertRaises(ProjectNotFound):
            delete_project(name='org/proj', api_token=API_TOKEN)

    def test_delete_project_permissions(self, swagger_client_factory):
        swagger_client = self._get_swagger_client_mock(swagger_client_factory)

        # when:
        swagger_client.api.deleteProject.side_effect = HTTPForbidden(response=MagicMock())

        # then:
        with self.assertRaises(AccessRevokedOnDeletion):
            delete_project(name='org/proj', api_token=API_TOKEN)

    def test_create_project_already_exists(self, swagger_client_factory):
        swagger_client = self._get_swagger_client_mock(swagger_client_factory)

        # given:
        organization = Mock(
            id=str(uuid.uuid4())
        )
        organization.name = 'org'
        organizations = [
            organization
        ]

        # when:
        swagger_client.api.listOrganizations.return_value.response = BravadoResponseMock(
            result=organizations,
        )
        swagger_client.api.createProject.side_effect = HTTPBadRequest(
            response=MagicMock(),
            swagger_result=MagicMock(
                code=None,
                errorType={'name': 'validationError'},
                message=None,
                title='Validation Errors',
                type=None,
                validationErrors=[{'path': ['name'], 'errors': [{'errorCode': {'name': 'ERR_NOT_UNIQUE'}}]}]
            )
        )

        # then:
        with self.assertRaises(ProjectAlreadyExists):
            create_project(name='org/proj', key='PRJ', api_token=API_TOKEN)

    def test_create_project_no_workspace(self, swagger_client_factory):
        swagger_client = self._get_swagger_client_mock(swagger_client_factory)

        # given:
        organization = Mock(
            id=str(uuid.uuid4())
        )
        organization.name = 'org'
        organizations = [
            organization
        ]

        # when:
        swagger_client.api.listOrganizations.return_value.response = BravadoResponseMock(
            result=organizations,
        )

        # then:
        with self.assertRaises(WorkspaceNotFound):
            create_project(name='not_an_org/proj', key='PRJ', api_token=API_TOKEN)

    def test_create_project(self, swagger_client_factory):
        swagger_client = self._get_swagger_client_mock(swagger_client_factory)

        # given:
        organization = Mock(
            id=str(uuid.uuid4())
        )
        organization.name = 'org'
        organizations = [
            organization
        ]

        # and:
        project = Mock(
            organizationName='org'
        )
        project.name = 'proj'

        # when:
        swagger_client.api.listOrganizations.return_value.response = BravadoResponseMock(
            result=organizations,
        )
        swagger_client.api.createProject.return_value.response = BravadoResponseMock(
            result=project,
        )

        # then:
        self.assertEqual('org/proj', create_project(name='org/proj', key='PRJ', api_token=API_TOKEN))

    def test_add_project_member_project_not_found(self, swagger_client_factory):
        swagger_client = self._get_swagger_client_mock(swagger_client_factory)

        # when:
        swagger_client.api.addProjectMember.side_effect = HTTPNotFound(response=MagicMock())

        # then:
        with self.assertRaises(ProjectNotFound):
            add_project_member(name='org/proj', username='tester', role=MemberRole.VIEWER, api_token=API_TOKEN)

    def test_add_project_member_member_without_access(self, swagger_client_factory):
        swagger_client = self._get_swagger_client_mock(swagger_client_factory)

        # when:
        swagger_client.api.addProjectMember.side_effect = HTTPConflict(response=MagicMock())

        # then:
        with self.assertRaises(UserAlreadyHasAccess):
            add_project_member(name='org/proj', username='tester', role=MemberRole.VIEWER, api_token=API_TOKEN)

    def test_remove_project_member_project_not_found(self, swagger_client_factory):
        swagger_client = self._get_swagger_client_mock(swagger_client_factory)

        # when:
        swagger_client.api.deleteProjectMember.side_effect = HTTPNotFound(response=MagicMock())

        # then:
        with self.assertRaises(ProjectNotFound):
            remove_project_member(name='org/proj', username='tester', api_token=API_TOKEN)

    def test_remove_project_member_no_user(self, swagger_client_factory):
        swagger_client = self._get_swagger_client_mock(swagger_client_factory)

        # when:
        swagger_client.api.deleteProjectMember.side_effect = HTTPUnprocessableEntity(response=MagicMock())

        # then:
        with self.assertRaises(UserNotExistsOrWithoutAccess):
            remove_project_member(name='org/proj', username='tester', api_token=API_TOKEN)

    def test_remove_project_member_permissions(self, swagger_client_factory):
        swagger_client = self._get_swagger_client_mock(swagger_client_factory)

        # when:
        swagger_client.api.deleteProjectMember.side_effect = HTTPForbidden(response=MagicMock())

        # then:
        with self.assertRaises(AccessRevokedOnMemberRemoval):
            remove_project_member(name='org/proj', username='tester', api_token=API_TOKEN)



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import json
import logging
import threading
from json.decoder import JSONDecodeError

from typing import TYPE_CHECKING, Optional

import click
from websocket import WebSocketConnectionClosedException

from neptune.internal.websockets.reconnecting_websocket import ReconnectingWebsocket
from neptune.new.attributes.constants import SIGNAL_TYPE_ABORT, SIGNAL_TYPE_STOP, SYSTEM_FAILED_ATTRIBUTE_PATH

from neptune.new.internal.background_job import BackgroundJob
from neptune.new.internal.threading.daemon import Daemon
from neptune.new.internal.utils import process_killer
from neptune.new.internal.websockets.websockets_factory import WebsocketsFactory

if TYPE_CHECKING:
    from neptune.new.run import Run

_logger = logging.getLogger(__name__)


class WebsocketSignalsBackgroundJob(BackgroundJob):

    def __init__(self, ws_factory: WebsocketsFactory):
        self._ws_factory = ws_factory
        self._thread: 'Optional[WebsocketSignalsBackgroundJob._ListenerThread]' = None
        self._started = False

    def start(self, run: 'Run'):
        self._thread = self._ListenerThread(run, self._ws_factory.create())
        self._thread.start()
        self._started = True

    def stop(self):
        if not self._started:
            return
        self._thread.interrupt()
        self._thread.shutdown_ws_client()

    def join(self, seconds: Optional[float] = None):
        if not self._started or threading.get_ident() == self._thread.ident:
            return
        self._thread.join(seconds)
        # Just in case. There is possible race condition when connection can be reestablished after being shutdown.
        self._thread.shutdown_ws_client()

    class _ListenerThread(Daemon):

        def __init__(self, run: 'Run', ws_client: ReconnectingWebsocket):
            super().__init__(sleep_time=0)
            self._run = run
            self._ws_client = ws_client

        def work(self) -> None:
            try:
                raw_message = self._ws_client.recv()
                if raw_message is None or self._is_heartbeat(raw_message):
                    return
                else:
                    self._handler_message(raw_message)
            except WebSocketConnectionClosedException:
                pass

        def _handler_message(self, msg: str):
            try:
                json_msg = json.loads(msg)
                msg_type = json_msg.get("type")
                msg_body = json_msg.get("body")
                if not msg_type:
                    click.echo("Malformed websocket signal: missing type", err=True)
                    return
                if not isinstance(msg_type, str):
                    click.echo(f"Malformed websocket signal: type is {type(msg_type)}, should be str", err=True)
                    return
                if msg_type.lower() == SIGNAL_TYPE_STOP:
                    self._handle_stop(msg_body)
                elif msg_type.lower() == SIGNAL_TYPE_ABORT:
                    self._handle_abort(msg_body)
            except JSONDecodeError as ex:
                click.echo(f"Malformed websocket signal: {ex}, message: {msg}", err=True)

        def _handle_stop(self, msg_body):
            msg_body = msg_body or dict()
            if not isinstance(msg_body, dict):
                click.echo(f"Malformed websocket signal: body is {type(msg_body)}, should be dict", err=True)
                return
            run_id = self._run["sys/id"].fetch()
            click.echo(f"Run {run_id} received stop signal. Exiting", err=True)
            seconds = msg_body.get("seconds")
            self._run.stop(seconds=seconds)
            process_killer.kill_me()

        def _handle_abort(self, msg_body):
            msg_body = msg_body or dict()
            if not isinstance(msg_body, dict):
                click.echo(f"Malformed websocket signal: body is {type(msg_body)}, should be dict", err=True)
                return
            run_id = self._run["sys/id"].fetch()
            click.echo(f"Run {run_id} received abort signal. Exiting", err=True)
            seconds = msg_body.get("seconds")
            self._run[SYSTEM_FAILED_ATTRIBUTE_PATH] = True
            self._run.stop(seconds=seconds)
            process_killer.kill_me()

        def shutdown_ws_client(self):
            self._ws_client.shutdown()

        @staticmethod
        def _is_heartbeat(message: str):
            return message.strip() == ''



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import logging
import os
import re
import time
from collections import namedtuple
from http.client import NOT_FOUND
from io import StringIO
from itertools import groupby
from pathlib import Path
from typing import Dict, List

import six
from bravado.exception import HTTPNotFound
from neptune.internal.websockets.reconnecting_websocket_factory import ReconnectingWebsocketFactory

from neptune.api_exceptions import (ExperimentNotFound, ExperimentOperationErrors, PathInExperimentNotFound,
                                    ProjectNotFound)
from neptune.exceptions import DeleteArtifactUnsupportedInAlphaException, DownloadArtifactUnsupportedException, \
    DownloadArtifactsUnsupportedException, DownloadSourcesException, FileNotFound, \
    NeptuneException
from neptune.experiments import Experiment
from neptune.internal.api_clients.hosted_api_clients.hosted_leaderboard_api_client \
    import HostedNeptuneLeaderboardApiClient
from neptune.internal.channels.channels import ChannelNamespace, ChannelType, ChannelValueType
from neptune.internal.storage.storage_utils import normalize_file_name
from neptune.internal.utils.alpha_integration import (
    AlphaChannelDTO,
    AlphaChannelWithValueDTO,
    AlphaParameterDTO,
    AlphaPropertyDTO,
    channel_type_to_clear_operation,
    channel_type_to_operation,
    channel_value_type_to_operation,
    deprecated_img_to_alpha_image
)
from neptune.model import ChannelWithLastValue, LeaderboardEntry
from neptune.new import exceptions as alpha_exceptions
from neptune.new.attributes import constants as alpha_consts
from neptune.new.attributes.constants import MONITORING_TRACEBACK_ATTRIBUTE_PATH, SYSTEM_FAILED_ATTRIBUTE_PATH
from neptune.new.exceptions import ClientHttpError
from neptune.new.internal import operation as alpha_operation
from neptune.new.internal.backends import hosted_file_operations as alpha_hosted_file_operations
from neptune.new.internal.backends.api_model import AttributeType
from neptune.new.internal.backends.operation_api_name_visitor import \
    OperationApiNameVisitor as AlphaOperationApiNameVisitor
from neptune.new.internal.backends.operation_api_object_converter import \
    OperationApiObjectConverter as AlphaOperationApiObjectConverter
from neptune.new.internal.operation import AssignString, ConfigFloatSeries, LogFloats, AssignBool, LogStrings
from neptune.new.internal.utils import base64_decode, base64_encode, paths as alpha_path_utils
from neptune.new.internal.utils.paths import parse_path
from neptune.utils import assure_directory_exists, with_api_exceptions_handler

_logger = logging.getLogger(__name__)

LegacyExperiment = namedtuple(
    'LegacyExperiment',
    'shortId '
    'name '
    'timeOfCreation '
    'timeOfCompletion '
    'runningTime '
    'owner '
    'storageSize '
    'channelsSize '
    'tags '
    'description '
    'hostname '
    'state '
    'properties '
    'parameters')

LegacyLeaderboardEntry = namedtuple(
    'LegacyExperiment',
    'id '
    'organizationName '
    'projectName '
    'shortId '
    'name '
    'state '
    'timeOfCreation '
    'timeOfCompletion '
    'runningTime '
    'owner '
    'size '
    'tags '
    'description '
    'channelsLastValues '
    'parameters '
    'properties'
)


class HostedAlphaLeaderboardApiClient(HostedNeptuneLeaderboardApiClient):

    @with_api_exceptions_handler
    def create_experiment(self,
                          project,
                          name,
                          description,
                          params,
                          properties,
                          tags,
                          abortable,  # deprecated in alpha
                          monitored,  # deprecated in alpha
                          git_info,
                          hostname,
                          entrypoint,
                          notebook_id,
                          checkpoint_id):
        if not isinstance(name, six.string_types):
            raise ValueError("Invalid name {}, should be a string.".format(name))
        if not isinstance(description, six.string_types):
            raise ValueError("Invalid description {}, should be a string.".format(description))
        if not isinstance(params, dict):
            raise ValueError("Invalid params {}, should be a dict.".format(params))
        if not isinstance(properties, dict):
            raise ValueError("Invalid properties {}, should be a dict.".format(properties))
        if hostname is not None and not isinstance(hostname, six.string_types):
            raise ValueError("Invalid hostname {}, should be a string.".format(hostname))
        if entrypoint is not None and not isinstance(entrypoint, six.string_types):
            raise ValueError("Invalid entrypoint {}, should be a string.".format(entrypoint))

        git_info = {
            "commit": {
                "commitId": git_info.commit_id,
                "message": git_info.message,
                "authorName": git_info.author_name,
                "authorEmail": git_info.author_email,
                "commitDate": git_info.commit_date
            },
            "repositoryDirty": git_info.repository_dirty,
            "currentBranch": git_info.active_branch,
            "remotes": git_info.remote_urls
        } if git_info else None

        api_params = {
            "notebookId": notebook_id,
            "checkpointId": checkpoint_id,
            "projectIdentifier": str(project.internal_id),
            "cliVersion": self.client_lib_version,
            "gitInfo": git_info,
            "customId": None,
        }

        kwargs = {
            'experimentCreationParams': api_params,
            'X-Neptune-CliVersion': self.client_lib_version,
            '_request_options': {"headers": {"X-Neptune-LegacyClient": "true"}}
        }

        try:
            api_experiment = self.leaderboard_swagger_client.api.createExperiment(**kwargs).response().result
        except HTTPNotFound:
            raise ProjectNotFound(project_identifier=project.full_id)

        experiment = self._convert_to_experiment(api_experiment, project)
        # Initialize new experiment
        init_experiment_operations = self._get_init_experiment_operations(name,
                                                                          description,
                                                                          params,
                                                                          properties,
                                                                          tags,
                                                                          hostname,
                                                                          entrypoint)
        self._execute_operations(
            experiment=experiment,
            operations=init_experiment_operations,
        )
        return experiment

    def upload_source_code(self, experiment, source_target_pairs):
        dest_path = alpha_path_utils.parse_path(alpha_consts.SOURCE_CODE_FILES_ATTRIBUTE_PATH)
        file_globs = [source_path for source_path, target_path in source_target_pairs]
        upload_files_operation = alpha_operation.UploadFileSet(
            path=dest_path,
            file_globs=file_globs,
            reset=True,
        )
        self._execute_upload_operations_with_400_retry(experiment, upload_files_operation)

    @with_api_exceptions_handler
    def get_experiment(self, experiment_id):
        api_attributes = self._get_api_experiment_attributes(experiment_id)
        attributes = api_attributes.attributes
        system_attributes = api_attributes.systemAttributes

        return LegacyExperiment(
            shortId=system_attributes.shortId.value,
            name=system_attributes.name.value,
            timeOfCreation=system_attributes.creationTime.value,
            timeOfCompletion=None,
            runningTime=system_attributes.runningTime.value,
            owner=system_attributes.owner.value,
            storageSize=system_attributes.size.value,
            channelsSize=0,
            tags=system_attributes.tags.values,
            description=system_attributes.description.value,
            hostname=system_attributes.hostname.value if system_attributes.hostname else None,
            state="running" if system_attributes.state.value == "running" else "succeeded",
            properties=[
                AlphaPropertyDTO(attr) for attr in attributes
                if AlphaPropertyDTO.is_valid_attribute(attr)
            ],
            parameters=[
                AlphaParameterDTO(attr) for attr in attributes
                if AlphaParameterDTO.is_valid_attribute(attr)
            ]
        )

    def update_experiment(self, experiment, properties):
        raise NeptuneException("`update_experiment` shouldn't be called.")

    @with_api_exceptions_handler
    def set_property(self, experiment, key, value):
        """Save attribute casted to string under `alpha_consts.PROPERTIES_ATTRIBUTE_SPACE` namespace"""
        self._execute_operations(
            experiment=experiment,
            operations=[alpha_operation.AssignString(
                path=alpha_path_utils.parse_path(f'{alpha_consts.PROPERTIES_ATTRIBUTE_SPACE}{key}'),
                value=str(value),
            )],
        )

    @with_api_exceptions_handler
    def remove_property(self, experiment, key):
        self._remove_attribute(experiment, str_path=f'{alpha_consts.PROPERTIES_ATTRIBUTE_SPACE}{key}')

    @with_api_exceptions_handler
    def update_tags(self, experiment, tags_to_add, tags_to_delete):
        operations = [
            alpha_operation.AddStrings(
                path=alpha_path_utils.parse_path(alpha_consts.SYSTEM_TAGS_ATTRIBUTE_PATH),
                values=tags_to_add,
            ),
            alpha_operation.RemoveStrings(
                path=alpha_path_utils.parse_path(alpha_consts.SYSTEM_TAGS_ATTRIBUTE_PATH),
                values=tags_to_delete,
            )
        ]
        self._execute_operations(
            experiment=experiment,
            operations=operations,
        )

    def upload_experiment_source(self, experiment, data, progress_indicator):
        raise NeptuneException(
            'This function should never be called for alpha project.')

    @staticmethod
    def _get_channel_attribute_path(channel_name: str, channel_namespace: ChannelNamespace) -> str:
        if channel_namespace == ChannelNamespace.USER:
            prefix = alpha_consts.LOG_ATTRIBUTE_SPACE
        else:
            prefix = alpha_consts.MONITORING_ATTRIBUTE_SPACE
        return f'{prefix}{channel_name}'

    def _create_channel(
            self,
            experiment: Experiment,
            channel_id: str,
            channel_name: str,
            channel_type: ChannelType,
            channel_namespace: ChannelNamespace):
        """This function is responsible for creating 'fake' channels in alpha projects.

        Since channels are abandoned in alpha api, we're mocking them using empty logging operation."""

        operation = channel_type_to_operation(channel_type)

        log_empty_operation = operation(
            path=alpha_path_utils.parse_path(self._get_channel_attribute_path(channel_name, channel_namespace)),
            values=[],
        )  # this operation is used to create empty attribute
        self._execute_operations(
            experiment=experiment,
            operations=[log_empty_operation],
        )
        return ChannelWithLastValue(
            AlphaChannelWithValueDTO(
                channelId=channel_id,
                channelName=channel_name,
                channelType=channel_type.value,
                x=None,
                y=None
            )
        )

    @with_api_exceptions_handler
    def create_channel(self, experiment, name, channel_type) -> ChannelWithLastValue:
        channel_id = f'{alpha_consts.LOG_ATTRIBUTE_SPACE}{name}'
        return self._create_channel(experiment, channel_id,
                                    channel_name=name,
                                    channel_type=ChannelType(channel_type),
                                    channel_namespace=ChannelNamespace.USER)

    def _get_channels(self, experiment) -> List[AlphaChannelDTO]:
        try:
            return [
                AlphaChannelDTO(attr) for attr in self._get_attributes(experiment.internal_id)
                if AlphaChannelDTO.is_valid_attribute(attr)
            ]
        except HTTPNotFound:
            # pylint: disable=protected-access
            raise ExperimentNotFound(
                experiment_short_id=experiment.id, project_qualified_name=experiment._project.full_id)

    @with_api_exceptions_handler
    def get_channels(self, experiment) -> Dict[str, AlphaChannelDTO]:
        api_channels = [
            channel for channel in self._get_channels(experiment)
            # return channels from LOG_ATTRIBUTE_SPACE namespace only
            if channel.id.startswith(alpha_consts.LOG_ATTRIBUTE_SPACE)
        ]
        return {ch.name: ch for ch in api_channels}

    @with_api_exceptions_handler
    def create_system_channel(self, experiment, name, channel_type) -> ChannelWithLastValue:
        channel_id = f'{alpha_consts.MONITORING_ATTRIBUTE_SPACE}{name}'
        return self._create_channel(experiment, channel_id,
                                    channel_name=name,
                                    channel_type=ChannelType(channel_type),
                                    channel_namespace=ChannelNamespace.SYSTEM)

    @with_api_exceptions_handler
    def get_system_channels(self, experiment) -> Dict[str, AlphaChannelDTO]:
        return {
            channel.name: channel
            for channel in self._get_channels(experiment)
            if (channel.channelType == ChannelType.TEXT.value
                and channel.id.startswith(alpha_consts.MONITORING_ATTRIBUTE_SPACE))
        }

    @with_api_exceptions_handler
    def send_channels_values(self, experiment, channels_with_values):
        send_operations = []
        for channel_with_values in channels_with_values:
            channel_value_type = channel_with_values.channel_type
            operation = channel_value_type_to_operation(channel_value_type)

            if channel_value_type == ChannelValueType.IMAGE_VALUE:
                # IMAGE_VALUE requires minor data modification before it's sent
                data_transformer = deprecated_img_to_alpha_image
            else:
                # otherwise use identity function as transformer
                data_transformer = lambda e: e

            ch_values = [
                alpha_operation.LogSeriesValue(
                    value=data_transformer(ch_value.value),
                    step=ch_value.x,
                    ts=ch_value.ts,
                )
                for ch_value in channel_with_values.channel_values
            ]
            send_operations.append(operation(
                path=alpha_path_utils.parse_path(self._get_channel_attribute_path(
                    channel_with_values.channel_name,
                    channel_with_values.channel_namespace)),
                values=ch_values,
            ))

        self._execute_operations(experiment, send_operations)

    def mark_succeeded(self, experiment):
        pass

    def mark_failed(self, experiment, traceback):
        operations = []
        path = parse_path(SYSTEM_FAILED_ATTRIBUTE_PATH)
        traceback_values = [LogStrings.ValueType(val, step=None, ts=time.time()) for val in traceback.split("\n")]
        operations.append(AssignBool(path=path, value=True))
        operations.append(LogStrings(values=traceback_values, path=parse_path(MONITORING_TRACEBACK_ATTRIBUTE_PATH)))
        self._execute_operations(experiment, operations)

    def ping_experiment(self, experiment):
        try:
            self.leaderboard_swagger_client.api.ping(experimentId=str(experiment.internal_id)).response().result
        except HTTPNotFound:
            # pylint: disable=protected-access
            raise ExperimentNotFound(
                experiment_short_id=experiment.id, project_qualified_name=experiment._project.full_id)

    @staticmethod
    def _get_attribute_name_for_metric(resource_type, gauge_name, gauges_count) -> str:
        if gauges_count > 1:
            return "monitoring/{}_{}".format(resource_type, gauge_name).lower()
        return "monitoring/{}".format(resource_type).lower()

    @with_api_exceptions_handler
    def create_hardware_metric(self, experiment, metric):
        operations = []
        gauges_count = len(metric.gauges)
        for gauge in metric.gauges:
            path = parse_path(self._get_attribute_name_for_metric(metric.resource_type, gauge.name(), gauges_count))
            operations.append(ConfigFloatSeries(path, min=metric.min_value, max=metric.max_value, unit=metric.unit))
        self._execute_operations(experiment, operations)

    @with_api_exceptions_handler
    def send_hardware_metric_reports(self, experiment, metrics, metric_reports):
        operations = []
        metrics_by_name = {metric.name: metric for metric in metrics}
        for report in metric_reports:
            metric = metrics_by_name.get(report.metric.name)
            gauges_count = len(metric.gauges)
            for gauge_name, metric_values in groupby(report.values, lambda value: value.gauge_name):
                metric_values = list(metric_values)
                path = parse_path(self._get_attribute_name_for_metric(metric.resource_type, gauge_name, gauges_count))
                operations.append(LogFloats(
                    path,
                    [LogFloats.ValueType(value.value, step=None, ts=value.timestamp) for value in metric_values]))
        self._execute_operations(experiment, operations)

    def log_artifact(self, experiment, artifact, destination=None):
        if isinstance(artifact, str):
            if os.path.isfile(artifact):
                target_name = os.path.basename(artifact) if destination is None else destination
                dest_path = self._get_dest_and_ext(target_name)
                operation = alpha_operation.UploadFile(
                    path=dest_path,
                    ext="",
                    file_path=os.path.abspath(artifact),
                )
            elif os.path.isdir(artifact):
                for path, file_destination in self._log_dir_artifacts(artifact, destination):
                    self.log_artifact(experiment, path, file_destination)
                return
            else:
                raise FileNotFound(artifact)
        elif hasattr(artifact, 'read'):
            if not destination:
                raise ValueError("destination is required for IO streams")
            dest_path = self._get_dest_and_ext(destination)
            data = artifact.read()
            content = data.encode('utf-8') if isinstance(data, str) else data
            operation = alpha_operation.UploadFileContent(
                path=dest_path,
                ext="",
                file_content=base64_encode(content)
            )
        else:
            raise ValueError("Artifact must be a local path or an IO object")

        self._execute_upload_operations_with_400_retry(experiment, operation)

    @staticmethod
    def _get_dest_and_ext(target_name):
        qualified_target_name = f'{alpha_consts.ARTIFACT_ATTRIBUTE_SPACE}{target_name}'
        return alpha_path_utils.parse_path(normalize_file_name(qualified_target_name))

    def _log_dir_artifacts(self, directory_path, destination):
        directory_path = Path(directory_path)
        prefix = directory_path.name if destination is None else destination
        for path in directory_path.glob('**/*'):
            if path.is_file():
                relative_path = path.relative_to(directory_path)
                file_destination = prefix + '/' + str(relative_path)
                yield str(path), file_destination

    def delete_artifacts(self, experiment, path):
        try:
            self._remove_attribute(experiment, str_path=f'{alpha_consts.ARTIFACT_ATTRIBUTE_SPACE}{path}')
        except ExperimentOperationErrors as e:
            if all(isinstance(err, alpha_exceptions.MetadataInconsistency) for err in e.errors):
                raise DeleteArtifactUnsupportedInAlphaException(path, experiment) from None
            raise

    @with_api_exceptions_handler
    def download_data(self, experiment: Experiment, path: str, destination):
        project_storage_path = f"artifacts/{path}"
        with self._download_raw_data(api_method=self.leaderboard_swagger_client.api.downloadAttribute,
                                     headers={"Accept": "application/octet-stream"},
                                     path_params={},
                                     query_params={
                                         "experimentId": experiment.internal_id,
                                         "attribute": project_storage_path
                                     }) as response:
            if response.status_code == NOT_FOUND:
                raise PathInExperimentNotFound(
                    path=path,
                    exp_identifier=experiment.internal_id)
            else:
                response.raise_for_status()

            with open(destination, "wb") as f:
                for chunk in response.iter_content(chunk_size=10 * 1024 * 1024):
                    if chunk:
                        f.write(chunk)

    def download_sources(self, experiment: Experiment, path=None, destination_dir=None):
        if path is not None:
            # in alpha all source files stored as single FileSet must be downloaded at once
            raise DownloadSourcesException(experiment)
        path = alpha_consts.SOURCE_CODE_FILES_ATTRIBUTE_PATH

        destination_dir = assure_directory_exists(destination_dir)

        download_request = self._get_file_set_download_request(
            experiment.internal_id,
            path)
        alpha_hosted_file_operations.download_file_set_attribute(
            swagger_client=self.leaderboard_swagger_client,
            download_id=download_request.id,
            destination=destination_dir)

    @with_api_exceptions_handler
    def _get_file_set_download_request(self, run_id: str, path: str):
        params = {
            'experimentId': run_id,
            'attribute': path,
        }
        return self.leaderboard_swagger_client.api.prepareForDownloadFileSetAttributeZip(**params).response().result

    def download_artifacts(self, experiment: Experiment, path=None, destination_dir=None):
        raise DownloadArtifactsUnsupportedException(experiment)

    def download_artifact(self, experiment: Experiment, path=None, destination_dir=None):
        destination_dir = assure_directory_exists(destination_dir)
        destination_path = os.path.join(destination_dir, os.path.basename(path))

        try:
            self.download_data(experiment, path, destination_path)
        except PathInExperimentNotFound:
            raise DownloadArtifactUnsupportedException(path, experiment) from None

    def _get_attributes(self, experiment_id) -> list:
        return self._get_api_experiment_attributes(experiment_id).attributes

    def _get_api_experiment_attributes(self, experiment_id):
        params = {
            'experimentId': experiment_id,
        }
        return self.leaderboard_swagger_client.api.getExperimentAttributes(**params).response().result

    def _remove_attribute(self, experiment, str_path: str):
        """Removes given attribute"""
        self._execute_operations(
            experiment=experiment,
            operations=[alpha_operation.DeleteAttribute(
                path=alpha_path_utils.parse_path(str_path),
            )],
        )

    @staticmethod
    def _get_client_config_args(api_token):
        return dict(
            X_Neptune_Api_Token=api_token,
            alpha="true",
        )

    def _execute_upload_operation(self,
                                  experiment: Experiment,
                                  upload_operation: alpha_operation.Operation):
        experiment_id = experiment.internal_id
        try:
            if isinstance(upload_operation, alpha_operation.UploadFile):
                alpha_hosted_file_operations.upload_file_attribute(
                    swagger_client=self.leaderboard_swagger_client,
                    run_id=experiment_id,
                    attribute=alpha_path_utils.path_to_str(upload_operation.path),
                    source=upload_operation.file_path,
                    ext=upload_operation.ext)
            elif isinstance(upload_operation, alpha_operation.UploadFileContent):
                alpha_hosted_file_operations.upload_file_attribute(
                    swagger_client=self.leaderboard_swagger_client,
                    run_id=experiment_id,
                    attribute=alpha_path_utils.path_to_str(upload_operation.path),
                    source=base64_decode(upload_operation.file_content),
                    ext=upload_operation.ext)
            elif isinstance(upload_operation, alpha_operation.UploadFileSet):
                alpha_hosted_file_operations.upload_file_set_attribute(
                    swagger_client=self.leaderboard_swagger_client,
                    run_id=experiment_id,
                    attribute=alpha_path_utils.path_to_str(upload_operation.path),
                    file_globs=upload_operation.file_globs,
                    reset=upload_operation.reset)
            else:
                raise NeptuneException("Upload operation in neither File or FileSet")
        except alpha_exceptions.NeptuneException as e:
            raise NeptuneException(e) from e

        return None

    def _execute_upload_operations_with_400_retry(
            self,
            experiment: Experiment,
            upload_operation: alpha_operation.Operation):
        while True:
            try:
                return self._execute_upload_operation(experiment, upload_operation)
            except ClientHttpError as ex:
                if "Length of stream does not match given range" not in ex.response:
                    raise ex

    @with_api_exceptions_handler
    def _execute_operations(self, experiment: Experiment, operations: List[alpha_operation.Operation]):
        experiment_id = experiment.internal_id
        file_operations = (
            alpha_operation.UploadFile,
            alpha_operation.UploadFileContent,
            alpha_operation.UploadFileSet
        )
        if any(isinstance(op, file_operations) for op in operations):
            raise NeptuneException("File operations must be handled directly by `_execute_upload_operation`,"
                                   " not by `_execute_operations` function call.")

        kwargs = {
            'experimentId': experiment_id,
            'operations': [
                {
                    'path': alpha_path_utils.path_to_str(op.path),
                    AlphaOperationApiNameVisitor().visit(op): AlphaOperationApiObjectConverter().convert(op)
                }
                for op in operations
            ]
        }
        try:
            result = self.leaderboard_swagger_client.api.executeOperations(**kwargs).response().result
            errors = [
                alpha_exceptions.MetadataInconsistency(err.errorDescription)
                for err in result
            ]
            if errors:
                raise ExperimentOperationErrors(errors=errors)
            return None
        except HTTPNotFound as e:
            # pylint: disable=protected-access
            raise ExperimentNotFound(
                experiment_short_id=experiment.id,
                project_qualified_name=experiment._project.full_id
            ) from e

    def _get_init_experiment_operations(self,
                                        name,
                                        description,
                                        params,
                                        properties,
                                        tags,
                                        hostname,
                                        entrypoint) -> List[alpha_operation.Operation]:
        """Returns operations required to initialize newly created experiment"""
        init_operations = list()

        # Assign experiment name
        init_operations.append(alpha_operation.AssignString(
            path=alpha_path_utils.parse_path(alpha_consts.SYSTEM_NAME_ATTRIBUTE_PATH),
            value=name,
        ))
        # Assign experiment description
        init_operations.append(alpha_operation.AssignString(
            path=alpha_path_utils.parse_path(alpha_consts.SYSTEM_DESCRIPTION_ATTRIBUTE_PATH),
            value=description,
        ))
        # Assign experiment parameters
        for p_name, p_val in params.items():
            parameter_type, string_value = self._get_parameter_with_type(p_val)
            operation_cls = alpha_operation.AssignFloat if parameter_type == 'double' else alpha_operation.AssignString
            init_operations.append(operation_cls(
                path=alpha_path_utils.parse_path(f'{alpha_consts.PARAMETERS_ATTRIBUTE_SPACE}{p_name}'),
                value=string_value,
            ))
        # Assign experiment properties
        for p_key, p_val in properties.items():
            init_operations.append(AssignString(
                path=alpha_path_utils.parse_path(f'{alpha_consts.PROPERTIES_ATTRIBUTE_SPACE}{p_key}'),
                value=str(p_val),
            ))
        # Assign tags
        if tags:
            init_operations.append(alpha_operation.AddStrings(
                path=alpha_path_utils.parse_path(alpha_consts.SYSTEM_TAGS_ATTRIBUTE_PATH),
                values=set(tags),
            ))
        # Assign source hostname
        if hostname:
            init_operations.append(alpha_operation.AssignString(
                path=alpha_path_utils.parse_path(alpha_consts.SYSTEM_HOSTNAME_ATTRIBUTE_PATH),
                value=hostname,
            ))
        # Assign source entrypoint
        if entrypoint:
            init_operations.append(alpha_operation.AssignString(
                path=alpha_path_utils.parse_path(alpha_consts.SOURCE_CODE_ENTRYPOINT_ATTRIBUTE_PATH),
                value=entrypoint,
            ))

        return init_operations

    @with_api_exceptions_handler
    def reset_channel(self, experiment, channel_id, channel_name, channel_type):
        op = channel_type_to_clear_operation(ChannelType(channel_type))
        attr_path = self._get_channel_attribute_path(channel_name, ChannelNamespace.USER)
        self._execute_operations(
            experiment=experiment,
            operations=[op(path=alpha_path_utils.parse_path(attr_path))],
        )

    @with_api_exceptions_handler
    def _get_channel_tuples_from_csv(self, experiment, channel_attribute_path):
        try:
            csv = self.leaderboard_swagger_client.api.getFloatSeriesValuesCSV(
                experimentId=experiment.internal_id, attribute=channel_attribute_path
            ).response().incoming_response.text
            lines = csv.split('\n')[:-1]
            return [line.split(',') for line in lines]
        except HTTPNotFound:
            # pylint: disable=protected-access
            raise ExperimentNotFound(
                experiment_short_id=experiment.id, project_qualified_name=experiment._project.full_id)

    @with_api_exceptions_handler
    def get_channel_points_csv(self, experiment, channel_internal_id, channel_name):
        try:
            channel_attr_path = self._get_channel_attribute_path(channel_name, ChannelNamespace.USER)
            values = self._get_channel_tuples_from_csv(experiment, channel_attr_path)
            step_and_value = [val[0] + ',' + val[2] for val in values]
            csv = StringIO()
            for line in step_and_value:
                csv.write(line + "\n")
            csv.seek(0)
            return csv
        except HTTPNotFound:
            # pylint: disable=protected-access
            raise ExperimentNotFound(
                experiment_short_id=experiment.id, project_qualified_name=experiment._project.full_id)

    @with_api_exceptions_handler
    def get_metrics_csv(self, experiment):
        metric_channels = [
            channel for channel in self._get_channels(experiment)
            if (channel.channelType == ChannelType.NUMERIC.value
                and channel.id.startswith(alpha_consts.MONITORING_ATTRIBUTE_SPACE))
        ]
        data = {
            # val[1] + ',' + val[2] is timestamp,value
            ch.name: [val[1] + ',' + val[2] for val in self._get_channel_tuples_from_csv(experiment, ch.id)]
            for ch in metric_channels
        }
        values_count = max(len(values) for values in data.values())
        csv = StringIO()
        csv.write(','.join(["x_{name},y_{name}".format(name=ch.name) for ch in metric_channels]))
        csv.write("\n")
        for i in range(0, values_count):
            csv.write(','.join([data[ch.name][i] if i < len(data[ch.name]) else "," for ch in metric_channels]))
            csv.write("\n")
        csv.seek(0)
        return csv

    @with_api_exceptions_handler
    def get_leaderboard_entries(self, project,
                                entry_types=None,  # deprecated
                                ids=None,
                                states=None,
                                owners=None,
                                tags=None,
                                min_running_time=None):
        if states is not None:
            states = [state if state == "running" else "idle" for state in states]
        try:
            def get_portion(limit, offset):
                return self.leaderboard_swagger_client.api.getLeaderboard(
                    projectIdentifier=project.full_id,
                    shortId=ids, state=states, owner=owners, tags=tags,
                    tagsMode='and', minRunningTimeSeconds=min_running_time,
                    sortBy=['sys/id'], sortFieldType=['string'], sortDirection=['ascending'],
                    limit=limit, offset=offset
                ).response().result.entries

            return [LeaderboardEntry(self._to_leaderboard_entry_dto(e))
                    for e in self._get_all_items(get_portion, step=100)]
        except HTTPNotFound:
            raise ProjectNotFound(project_identifier=project.full_id)

    def websockets_factory(self, project_id, experiment_id):
        base_url = re.sub(r'^http', 'ws', self.api_address) + '/api/notifications/v1'
        return ReconnectingWebsocketFactory(
            backend=self,
            url=base_url + f'/runs/{project_id}/{experiment_id}/signal'
        )

    @staticmethod
    def _to_leaderboard_entry_dto(experiment_attributes):
        attributes = experiment_attributes.attributes
        system_attributes = experiment_attributes.systemAttributes

        def is_channel_namespace(name):
            return name.startswith(alpha_consts.LOG_ATTRIBUTE_SPACE) \
                   or name.startswith(alpha_consts.MONITORING_ATTRIBUTE_SPACE)

        numeric_channels = [
            HostedAlphaLeaderboardApiClient._float_series_to_channel_last_value_dto(attr)
            for attr in attributes
            if (attr.type == AttributeType.FLOAT_SERIES.value
                and is_channel_namespace(attr.name)
                and attr.floatSeriesProperties.last is not None)
        ]
        text_channels = [
            HostedAlphaLeaderboardApiClient._string_series_to_channel_last_value_dto(attr)
            for attr in attributes
            if (attr.type == AttributeType.STRING_SERIES.value
                and is_channel_namespace(attr.name)
                and attr.stringSeriesProperties.last is not None)
        ]

        return LegacyLeaderboardEntry(
            id=experiment_attributes.id,
            organizationName=experiment_attributes.organizationName,
            projectName=experiment_attributes.projectName,
            shortId=system_attributes.shortId.value,
            name=system_attributes.name.value,
            state="running" if system_attributes.state.value == "running" else "succeeded",
            timeOfCreation=system_attributes.creationTime.value,
            timeOfCompletion=None,
            runningTime=system_attributes.runningTime.value,
            owner=system_attributes.owner.value,
            size=system_attributes.size.value,
            tags=system_attributes.tags.values,
            description=system_attributes.description.value,
            channelsLastValues=numeric_channels + text_channels,
            parameters=[
                AlphaParameterDTO(attr) for attr in attributes
                if AlphaParameterDTO.is_valid_attribute(attr)
            ],
            properties=[
                AlphaPropertyDTO(attr) for attr in attributes
                if AlphaPropertyDTO.is_valid_attribute(attr)
            ]
        )

    @staticmethod
    def _float_series_to_channel_last_value_dto(attribute):
        return AlphaChannelWithValueDTO(
            channelId=attribute.name,
            channelName=attribute.name.split('/', 1)[-1],
            channelType="numeric",
            x=attribute.floatSeriesProperties.lastStep,
            y=attribute.floatSeriesProperties.last
        )

    @staticmethod
    def _string_series_to_channel_last_value_dto(attribute):
        return AlphaChannelWithValueDTO(
            channelId=attribute.name,
            channelName=attribute.name.split('/', 1)[-1],
            channelType="text",
            x=attribute.stringSeriesProperties.lastStep,
            y=attribute.stringSeriesProperties.last
        )



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# pylint: disable=protected-access
from mock import MagicMock, call

from neptune.new.internal.operation import ClearStringSet, AddStrings, RemoveStrings
from neptune.new.attributes.sets.string_set import StringSet, StringSetVal

from tests.neptune.new.attributes.test_attribute_base import TestAttributeBase


class TestStringSet(TestAttributeBase):

    def test_assign(self):
        value = StringSetVal(["ert", "qwe"])
        expected = {"ert", "qwe"}

        processor = MagicMock()
        exp, path, wait = self._create_run(processor), self._random_path(), self._random_wait()
        var = StringSet(exp, path)
        var.assign(value, wait=wait)
        self.assertEqual(2, processor.enqueue_operation.call_count)
        processor.enqueue_operation.assert_has_calls([
            call(ClearStringSet(path), False),
            call(AddStrings(path, expected), wait)
        ])

    def test_assign_empty(self):
        processor = MagicMock()
        exp, path, wait = self._create_run(processor), self._random_path(), self._random_wait()
        var = StringSet(exp, path)
        var.assign(StringSetVal([]), wait=wait)
        processor.enqueue_operation.assert_called_once_with(ClearStringSet(path), wait)

    def test_assign_type_error(self):
        values = [{5.}, {"text"}, {}, [5.], ["text"], [], 55, "string", None]
        for value in values:
            with self.assertRaises(TypeError):
                StringSet(MagicMock(), MagicMock()).assign(value)

    def test_add(self):
        processor = MagicMock()
        exp, path, wait = self._create_run(processor), self._random_path(), self._random_wait()
        var = StringSet(exp, path)
        var.add(["a", "bb", "ccc"], wait=wait)
        processor.enqueue_operation.assert_called_once_with(AddStrings(path, {"a", "bb", "ccc"}), wait)

    def test_add_single_value(self):
        processor = MagicMock()
        exp, path, wait = self._create_run(processor), self._random_path(), self._random_wait()
        var = StringSet(exp, path)
        var.add("ccc", wait=wait)
        processor.enqueue_operation.assert_called_once_with(AddStrings(path, {"ccc"}), wait)

    def test_remove(self):
        processor = MagicMock()
        exp, path, wait = self._create_run(processor), self._random_path(), self._random_wait()
        var = StringSet(exp, path)
        var.remove(["a", "bb", "ccc"], wait=wait)
        processor.enqueue_operation.assert_called_once_with(RemoveStrings(path, {"a", "bb", "ccc"}), wait)

    def test_remove_single_value(self):
        processor = MagicMock()
        exp, path, wait = self._create_run(processor), self._random_path(), self._random_wait()
        var = StringSet(exp, path)
        var.remove("bb", wait=wait)
        processor.enqueue_operation.assert_called_once_with(RemoveStrings(path, {"bb"}), wait)

    def test_clear(self):
        processor = MagicMock()
        exp, path, wait = self._create_run(processor), self._random_path(), self._random_wait()
        var = StringSet(exp, path)
        var.clear(wait=wait)
        processor.enqueue_operation.assert_called_once_with(ClearStringSet(path), wait)

    def test_get(self):
        exp, path = self._create_run(), self._random_path()
        var = StringSet(exp, path)
        var.add(["abc", "xyz"])
        var.remove(["abc"])
        var.add(["hej", "lol"])
        self.assertEqual({"xyz", "hej", "lol"}, var.fetch())



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging
import threading
from functools import wraps
from typing import Dict, TYPE_CHECKING

from neptune.model import ChannelWithLastValue

from neptune.backend import LeaderboardApiClient

from neptune.exceptions import ProjectMigratedToNewStructure
from neptune.internal.api_clients.hosted_api_clients.hosted_leaderboard_api_client import \
    HostedNeptuneLeaderboardApiClient

if TYPE_CHECKING:
    from neptune.internal.api_clients import HostedNeptuneBackendApiClient

_logger = logging.getLogger(__name__)


# pylint: disable=protected-access
def with_migration_handling(func):
    @wraps(func)
    def wrapper(api_proxy: 'MigrationSwitchLeaderboardApiClientProxy', *args, **kwargs):
        try:
            return func(api_proxy, *args, **kwargs)
        except ProjectMigratedToNewStructure:
            if not api_proxy._switched:
                api_proxy._lock.acquire()
                if not api_proxy._switched:
                    api_proxy._client = api_proxy._backend_client.get_new_leaderboard_client()
                    api_proxy._switched = True
                api_proxy._lock.release()
            return func(api_proxy, *args, **kwargs)

    return wrapper


class MigrationSwitchLeaderboardApiClientProxy(LeaderboardApiClient):

    def __init__(self, api_client: HostedNeptuneLeaderboardApiClient, backend_client: 'HostedNeptuneBackendApiClient'):
        self._client = api_client
        self._backend_client = backend_client
        self._lock = threading.RLock()
        self._switched = False

    @property
    def http_client(self):
        return self._client.http_client

    @property
    def backend_client(self):
        return self._client.backend_client

    @property
    def authenticator(self):
        return self._client.authenticator

    @property
    def credentials(self):
        return self._client.credentials

    @property
    def backend_swagger_client(self):
        return self._client.backend_swagger_client

    @property
    def client_lib_version(self):
        return self._client.client_lib_version

    @property
    def api_address(self):
        return self._client.api_address

    @property
    def display_address(self):
        return self._client.display_address

    @property
    def proxies(self):
        return self._client.proxies

    @with_migration_handling
    def get_project_members(self, project_identifier):
        return self._client.get_project_members(project_identifier)

    @with_migration_handling
    def get_leaderboard_entries(self, project,
                                entry_types=None, ids=None,
                                states=None, owners=None, tags=None,
                                min_running_time=None):
        return self._client.get_leaderboard_entries(project, entry_types, ids, states, owners, tags, min_running_time)

    @with_migration_handling
    def websockets_factory(self, project_id, experiment_id):
        return self._client.websockets_factory(project_id, experiment_id)

    @with_migration_handling
    def get_channel_points_csv(self, experiment, channel_internal_id, channel_name):
        return self._client.get_channel_points_csv(experiment, channel_internal_id, channel_name)

    @with_migration_handling
    def get_metrics_csv(self, experiment):
        return self._client.get_metrics_csv(experiment)

    @with_migration_handling
    def create_experiment(self, project, name, description,
                          params, properties, tags, abortable,
                          monitored, git_info, hostname, entrypoint,
                          notebook_id, checkpoint_id):
        return self._client.create_experiment(
            project, name, description,
            params, properties, tags, abortable,
            monitored, git_info, hostname, entrypoint,
            notebook_id, checkpoint_id)

    @with_migration_handling
    def upload_source_code(self, experiment, source_target_pairs):
        return self._client.upload_source_code(experiment, source_target_pairs)

    @with_migration_handling
    def get_notebook(self, project, notebook_id):
        return self._client.get_notebook(project, notebook_id)

    @with_migration_handling
    def get_last_checkpoint(self, project, notebook_id):
        return self._client.get_last_checkpoint(project, notebook_id)

    @with_migration_handling
    def create_notebook(self, project):
        return self._client.create_notebook(project)

    @with_migration_handling
    def create_checkpoint(self, notebook_id, jupyter_path, _file=None):
        return self._client.create_checkpoint(notebook_id, jupyter_path, _file)

    @with_migration_handling
    def get_experiment(self, experiment_id):
        return self._client.get_experiment(experiment_id)

    @with_migration_handling
    def update_experiment(self, experiment, properties):
        return self._client.update_experiment(experiment, properties)

    @with_migration_handling
    def set_property(self, experiment, key, value):
        return self._client.set_property(experiment, key, value)

    @with_migration_handling
    def remove_property(self, experiment, key):
        return self._client.remove_property(experiment, key)

    @with_migration_handling
    def update_tags(self, experiment, tags_to_add, tags_to_delete):
        return self._client.update_tags(experiment, tags_to_add, tags_to_delete)

    @with_migration_handling
    def upload_experiment_source(self, experiment, data, progress_indicator):
        return self._client.upload_experiment_source(experiment, data, progress_indicator)

    @with_migration_handling
    def extract_experiment_source(self, experiment, data):
        return self._client.extract_experiment_source(experiment, data)

    @with_migration_handling
    def create_channel(self, experiment, name, channel_type) -> ChannelWithLastValue:
        return self._client.create_channel(experiment, name, channel_type)

    @with_migration_handling
    def get_channels(self, experiment) -> Dict[str, object]:
        return self._client.get_channels(experiment)

    @with_migration_handling
    def reset_channel(self, experiment, channel_id, channel_name, channel_type):
        return self._client.reset_channel(experiment, channel_id, channel_name, channel_type)

    @with_migration_handling
    def create_system_channel(self, experiment, name, channel_type) -> ChannelWithLastValue:
        return self._client.create_system_channel(experiment, name, channel_type)

    @with_migration_handling
    def get_system_channels(self, experiment) -> Dict[str, object]:
        return self._client.get_system_channels(experiment)

    @with_migration_handling
    def send_channels_values(self, experiment, channels_with_values):
        return self._client.send_channels_values(experiment, channels_with_values)

    @with_migration_handling
    def mark_succeeded(self, experiment):
        return self._client.mark_succeeded(experiment)

    @with_migration_handling
    def mark_failed(self, experiment, traceback):
        return self._client.mark_failed(experiment, traceback)

    @with_migration_handling
    def ping_experiment(self, experiment):
        return self._client.ping_experiment(experiment)

    @with_migration_handling
    def create_hardware_metric(self, experiment, metric):
        return self._client.create_hardware_metric(experiment, metric)

    @with_migration_handling
    def send_hardware_metric_reports(self, experiment, metrics, metric_reports):
        return self._client.send_hardware_metric_reports(experiment, metrics, metric_reports)

    @with_migration_handling
    def log_artifact(self, experiment, artifact, destination=None):
        return self._client.log_artifact(experiment, artifact, destination)

    @with_migration_handling
    def delete_artifacts(self, experiment, path):
        return self._client.delete_artifacts(experiment, path)

    @with_migration_handling
    def download_data(self, experiment, path, destination):
        return self._client.download_data(experiment, path, destination)

    @with_migration_handling
    def download_sources(self, experiment, path=None, destination_dir=None):
        return self._client.download_sources(experiment, path, destination_dir)

    @with_migration_handling
    def download_artifacts(self, experiment, path=None, destination_dir=None):
        return self._client.download_artifacts(experiment, path, destination_dir)

    @with_migration_handling
    def download_artifact(self, experiment, path=None, destination_dir=None):
        return self._client.download_artifact(experiment, path, destination_dir)



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import contextlib
from tempfile import NamedTemporaryFile
import os


@contextlib.contextmanager
def create_file(content=None, binary_mode=False) -> str:
    """
    A lot this is motivated by:
    Whether the name can be used to open the file a second time,
    while the named temporary file is still open, varies across platforms
    (it can be so used on Unix; it cannot on Windows NT or later).
     ref. https://docs.python.org/3.9/library/tempfile.html#tempfile.NamedTemporaryFile
    """
    if binary_mode:
        mode = 'wb'
    else:
        mode = 'w'
    with NamedTemporaryFile(mode, delete=False) as file:
        if content:
            file.write(content)
        file.close()
        yield file.name
        os.remove(file.name)



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import atexit
import logging
import os
import os.path
import threading
from platform import node as get_hostname

import click
import pandas as pd
import six

from neptune.envs import NOTEBOOK_ID_ENV_NAME, NOTEBOOK_PATH_ENV_NAME
from neptune.exceptions import NeptuneNoExperimentContextException
from neptune.experiments import Experiment
from neptune.internal.abort import DefaultAbortImpl
from neptune.internal.notebooks.notebooks import create_checkpoint
from neptune.internal.utils.source_code import get_source_code_to_upload
from neptune.utils import as_list, map_keys, get_git_info, discover_git_repo_location

_logger = logging.getLogger(__name__)


class Project(object):
    # pylint: disable=redefined-builtin

    """A class for storing information and managing Neptune project.

    Args:
        backend (:class:`~neptune.ApiClient`, required): A ApiClient object.
        internal_id (:obj:`str`, required): ID of the project.
        namespace (:obj:`str`, required): It can either be your workspace or user name.
        name (:obj:`str`, required): project name.

    Note:
        ``namespace`` and ``name`` joined together with ``/`` form ``project_qualified_name``.
    """

    def __init__(self, backend, internal_id, namespace, name):
        self._backend = backend
        self.internal_id = internal_id
        self.namespace = namespace
        self.name = name

        self._experiments_stack = []
        self.__lock = threading.RLock()
        atexit.register(self._shutdown_hook)

    def get_members(self):
        """Retrieve a list of project members.

        Returns:
            :obj:`list` of :obj:`str` - A list of usernames of project members.

        Examples:

            .. code:: python3

                project = session.get_projects('neptune-ai')['neptune-ai/Salt-Detection']
                project.get_members()

        """
        project_members = self._backend.get_project_members(self.internal_id)
        return [member.registeredMemberInfo.username for member in project_members if member.registeredMemberInfo]

    def get_experiments(self, id=None, state=None, owner=None, tag=None, min_running_time=None):
        """Retrieve list of experiments matching the specified criteria.

        All parameters are optional, each of them specifies a single criterion.
        Only experiments matching all of the criteria will be returned.

        Args:
            id (:obj:`str` or :obj:`list` of :obj:`str`, optional, default is ``None``):
                | An experiment id like ``'SAN-1'`` or list of ids like ``['SAN-1', 'SAN-2']``.
                | Matching any element of the list is sufficient to pass criterion.
            state (:obj:`str` or :obj:`list` of :obj:`str`, optional, default is ``None``):
                | An experiment state like ``'succeeded'`` or list of states like ``['succeeded', 'running']``.
                | Possible values: ``'running'``, ``'succeeded'``, ``'failed'``, ``'aborted'``.
                | Matching any element of the list is sufficient to pass criterion.
            owner (:obj:`str` or :obj:`list` of :obj:`str`, optional, default is ``None``):
                | *Username* of the experiment owner (User who created experiment is an owner) like ``'josh'``
                  or list of owners like ``['frederic', 'josh']``.
                | Matching any element of the list is sufficient to pass criterion.
            tag (:obj:`str` or :obj:`list` of :obj:`str`, optional, default is ``None``):
                 | An experiment tag like ``'lightGBM'`` or list of tags like ``['pytorch', 'cycleLR']``.
                 | Only experiments that have all specified tags will match this criterion.
            min_running_time (:obj:`int`, optional, default is ``None``):
                Minimum running time of an experiment in seconds, like ``2000``.

        Returns:
            :obj:`list` of :class:`~neptune.experiments.Experiment` objects.

        Examples:

            .. code:: python3

                # Fetch a project
                project = session.get_projects('neptune-ai')['neptune-ai/Salt-Detection']

                # Get list of experiments
                project.get_experiments(state=['aborted'], owner=['neyo'], min_running_time=100000)

                # Example output:
                # [Experiment(SAL-1609),
                #  Experiment(SAL-1765),
                #  Experiment(SAL-1941),
                #  Experiment(SAL-1960),
                #  Experiment(SAL-2025)]
        """
        leaderboard_entries = self._fetch_leaderboard(id, state, owner, tag, min_running_time)
        return [
            Experiment(self._backend, self, entry.id, entry.internal_id)
            for entry in leaderboard_entries
        ]

    def get_leaderboard(self, id=None, state=None, owner=None, tag=None, min_running_time=None):
        """Fetch Neptune experiments view as pandas ``DataFrame``.

        **returned DataFrame**

        | In the returned ``DataFrame`` each *row* is an experiment and *columns* represent all system properties,
          numeric and text logs, parameters and properties in these experiments.
        | Note that, returned ``DataFrame`` does not contain all columns across the entire project.
        | Some columns may be empty, since experiments may define various logs, properties, etc.
        | For each log at most one (the last one) value is returned per experiment.
        | Text values are trimmed to 255 characters.

        **about parameters**

        All parameters are optional, each of them specifies a single criterion.
        Only experiments matching all of the criteria will be returned.

        Args:
            id (:obj:`str` or :obj:`list` of :obj:`str`, optional, default is ``None``):
                | An experiment id like ``'SAN-1'`` or list of ids like ``['SAN-1', 'SAN-2']``.
                | Matching any element of the list is sufficient to pass criterion.
            state (:obj:`str` or :obj:`list` of :obj:`str`, optional, default is ``None``):
                | An experiment state like ``'succeeded'`` or list of states like ``['succeeded', 'running']``.
                | Possible values: ``'running'``, ``'succeeded'``, ``'failed'``, ``'aborted'``.
                | Matching any element of the list is sufficient to pass criterion.
            owner (:obj:`str` or :obj:`list` of :obj:`str`, optional, default is ``None``):
                | *Username* of the experiment owner (User who created experiment is an owner) like ``'josh'``
                  or list of owners like ``['frederic', 'josh']``.
                | Matching any element of the list is sufficient to pass criterion.
            tag (:obj:`str` or :obj:`list` of :obj:`str`, optional, default is ``None``):
                | An experiment tag like ``'lightGBM'`` or list of tags like ``['pytorch', 'cycleLR']``.
                | Only experiments that have all specified tags will match this criterion.
            min_running_time (:obj:`int`, optional, default is ``None``):
                Minimum running time of an experiment in seconds, like ``2000``.

        Returns:
            :obj:`pandas.DataFrame` - Fetched Neptune experiments view.

        Examples:

            .. code:: python3

                # Fetch a project.
                project = session.get_projects('neptune-ai')['neptune-ai/Salt-Detection']

                # Get DataFrame that resembles experiment view.
                project.get_leaderboard(state=['aborted'], owner=['neyo'], min_running_time=100000)
        """

        leaderboard_entries = self._fetch_leaderboard(id, state, owner, tag, min_running_time)

        def make_row(entry):
            channels = dict(
                ('channel_{}'.format(ch.name), ch.trimmed_y) for ch in entry.channels
            )

            parameters = map_keys('parameter_{}'.format, entry.parameters)
            properties = map_keys('property_{}'.format, entry.properties)

            r = {}
            r.update(entry.system_properties)
            r.update(channels)
            r.update(parameters)
            r.update(properties)
            return r

        rows = ((n, make_row(e)) for (n, e) in enumerate(leaderboard_entries))

        df = pd.DataFrame.from_dict(data=dict(rows), orient='index')
        df = df.reindex(self._sort_leaderboard_columns(df.columns), axis='columns')
        return df

    def create_experiment(self,
                          name=None,
                          description=None,
                          params=None,
                          properties=None,
                          tags=None,
                          upload_source_files=None,
                          abort_callback=None,
                          logger=None,
                          upload_stdout=True,
                          upload_stderr=True,
                          send_hardware_metrics=True,
                          run_monitoring_thread=True,
                          handle_uncaught_exceptions=True,
                          git_info=None,
                          hostname=None,
                          notebook_id=None,
                          notebook_path=None):
        """Create and start Neptune experiment.

        Create experiment, set its status to `running` and append it to the top of the experiments view.
        All parameters are optional, hence minimal invocation: ``neptune.create_experiment()``.

        Args:
            name (:obj:`str`, optional, default is ``'Untitled'``):
                Editable name of the experiment.
                Name is displayed in the experiment's `Details` (`Metadata` section)
                and in `experiments view` as a column.

            description (:obj:`str`, optional, default is ``''``):
                Editable description of the experiment.
                Description is displayed in the experiment's `Details` (`Metadata` section)
                and can be displayed in the `experiments view` as a column.

            params (:obj:`dict`, optional, default is ``{}``):
                Parameters of the experiment.
                After experiment creation ``params`` are read-only
                (see: :meth:`~neptune.experiments.Experiment.get_parameters`).
                Parameters are displayed in the experiment's `Details` (`Parameters` section)
                and each key-value pair can be viewed in `experiments view` as a column.

            properties (:obj:`dict`, optional, default is ``{}``):
                Properties of the experiment.
                They are editable after experiment is created.
                Properties are displayed in the experiment's `Details` (`Properties` section)
                and each key-value pair can be viewed in `experiments view` as a column.

            tags (:obj:`list`, optional, default is ``[]``):
                Must be list of :obj:`str`. Tags of the experiment.
                They are editable after experiment is created
                (see: :meth:`~neptune.experiments.Experiment.append_tag`
                and :meth:`~neptune.experiments.Experiment.remove_tag`).
                Tags are displayed in the experiment's `Details` (`Metadata` section)
                and can be viewed in `experiments view` as a column.

            upload_source_files (:obj:`list` or :obj:`str`, optional, default is ``None``):
                List of source files to be uploaded. Must be list of :obj:`str` or single :obj:`str`.
                Uploaded sources are displayed in the experiment's `Source code` tab.

                | If ``None`` is passed, Python file from which experiment was created will be uploaded.
                | Pass empty list (``[]``) to upload no files.
                | Unix style pathname pattern expansion is supported. For example, you can pass ``'*.py'`` to upload
                  all python source files from the current directory.
                  For Python 3.5 or later, paths of uploaded files on server are resolved as relative to the
                | calculated common root of all uploaded source  files. For older Python versions, paths on server are
                | resolved always as relative to the current directory.
                  For recursion lookup use ``'**/*.py'`` (for Python 3.5 and later).
                  For more information see `glob library <https://docs.python.org/3/library/glob.html>`_.

            abort_callback (:obj:`callable`, optional, default is ``None``):
                Callback that defines how `abort experiment` action in the Web application should work.
                Actual behavior depends on your setup:

                    * (default) If ``abort_callback=None`` and `psutil <https://psutil.readthedocs.io/en/latest/>`_
                      is installed, then current process and it's children are aborted by sending `SIGTERM`.
                      If, after grace period, processes are not terminated, `SIGKILL` is sent.
                    * If ``abort_callback=None`` and `psutil <https://psutil.readthedocs.io/en/latest/>`_
                      is **not** installed, then `abort experiment` action just marks experiment as *aborted*
                      in the Web application. No action is performed on the current process.
                    * If ``abort_callback=callable``, then ``callable`` is executed when `abort experiment` action
                      in the Web application is triggered.

            logger (:obj:`logging.Logger` or `None`, optional, default is ``None``):
                If Python's `Logger <https://docs.python.org/3/library/logging.html#logging.Logger>`_
                is passed, new experiment's `text log`
                (see: :meth:`~neptune.experiments.Experiment.log_text`) with name `"logger"` is created.
                Each time `Python logger` logs new data, it is automatically sent to the `"logger"` in experiment.
                As a results all data from `Python logger` are in the `Logs` tab in the experiment.

            upload_stdout (:obj:`Boolean`, optional, default is ``True``):
                Whether to send stdout to experiment's *Monitoring*.

            upload_stderr (:obj:`Boolean`, optional, default is ``True``):
                Whether to send stderr to experiment's *Monitoring*.

            send_hardware_metrics (:obj:`Boolean`, optional, default is ``True``):
                Whether to send hardware monitoring logs (CPU, GPU, Memory utilization) to experiment's *Monitoring*.

            run_monitoring_thread (:obj:`Boolean`, optional, default is ``True``):
                Whether to run thread that pings Neptune server in order to determine if experiment is responsive.

            handle_uncaught_exceptions (:obj:`Boolean`, optional, default is ``True``):
                Two options ``True`` and ``False`` are possible:

                    * If set to ``True`` and uncaught exception occurs, then Neptune automatically place
                      `Traceback` in the experiment's `Details` and change experiment status to `Failed`.
                    * If set to ``False`` and uncaught exception occurs, then no action is performed
                      in the Web application. As a consequence, experiment's status is `running` or `not responding`.

            git_info (:class:`~neptune.git_info.GitInfo`, optional, default is ``None``):

                | Instance of the class :class:`~neptune.git_info.GitInfo` that provides information about
                  the git repository from which experiment was started.
                | If ``None`` is passed,
                  system attempts to automatically extract information about git repository in the following way:

                      * System looks for `.git` file in the current directory and, if not found,
                        goes up recursively until `.git` file will be found
                        (see: :meth:`~neptune.utils.get_git_info`).
                      * If there is no git repository,
                        then no information about git is displayed in experiment details in Neptune web application.

            hostname (:obj:`str`, optional, default is ``None``):
                If ``None``, neptune automatically get `hostname` information.
                User can also set `hostname` directly by passing :obj:`str`.

        Returns:
            :class:`~neptune.experiments.Experiment` object that is used to manage experiment and log data to it.

        Raises:
            `ExperimentValidationError`: When provided arguments are invalid.
            `ExperimentLimitReached`: When experiment limit in the project has been reached.

        Examples:

            .. code:: python3

                # minimal invoke
                neptune.create_experiment()

                # explicitly return experiment object
                experiment = neptune.create_experiment()

                # create experiment with name and two parameters
                neptune.create_experiment(name='first-pytorch-ever',
                                          params={'lr': 0.0005,
                                                  'dropout': 0.2})

                # create experiment with name and description, and no sources files uploaded
                neptune.create_experiment(name='neural-net-mnist',
                                          description='neural net trained on MNIST',
                                          upload_source_files=[])

                # Send all py files in cwd (excluding hidden files with names beginning with a dot)
                neptune.create_experiment(upload_source_files='*.py')

                # Send all py files from all subdirectories (excluding hidden files with names beginning with a dot)
                # Supported on Python 3.5 and later.
                neptune.create_experiment(upload_source_files='**/*.py')

                # Send all files and directories in cwd (excluding hidden files with names beginning with a dot)
                neptune.create_experiment(upload_source_files='*')

                # Send all files and directories in cwd including hidden files
                neptune.create_experiment(upload_source_files=['*', '.*'])

                # Send files with names being a single character followed by '.py' extension.
                neptune.create_experiment(upload_source_files='?.py')

                # larger example
                neptune.create_experiment(name='first-pytorch-ever',
                                          params={'lr': 0.0005,
                                                  'dropout': 0.2},
                                          properties={'key1': 'value1',
                                                      'key2': 17,
                                                      'key3': 'other-value'},
                                          description='write longer description here',
                                          tags=['list-of', 'tags', 'goes-here', 'as-list-of-strings'],
                                          upload_source_files=['training_with_pytorch.py', 'net.py'])
        """

        if name is None:
            name = "Untitled"

        if description is None:
            description = ""

        if params is None:
            params = {}

        if properties is None:
            properties = {}

        if tags is None:
            tags = []

        if git_info is None:
            git_info = get_git_info(discover_git_repo_location())

        if hostname is None:
            hostname = get_hostname()

        if notebook_id is None and os.getenv(NOTEBOOK_ID_ENV_NAME, None) is not None:
            notebook_id = os.environ[NOTEBOOK_ID_ENV_NAME]

        if isinstance(upload_source_files, six.string_types):
            upload_source_files = [upload_source_files]

        entrypoint, source_target_pairs = get_source_code_to_upload(upload_source_files=upload_source_files)

        if notebook_path is None and os.getenv(NOTEBOOK_PATH_ENV_NAME, None) is not None:
            notebook_path = os.environ[NOTEBOOK_PATH_ENV_NAME]

        abortable = abort_callback is not None or DefaultAbortImpl.requirements_installed()

        checkpoint_id = None
        if notebook_id is not None and notebook_path is not None:
            checkpoint = create_checkpoint(backend=self._backend,
                                           notebook_id=notebook_id,
                                           notebook_path=notebook_path)
            if checkpoint is not None:
                checkpoint_id = checkpoint.id

        experiment = self._backend.create_experiment(
            project=self,
            name=name,
            description=description,
            params=params,
            properties=properties,
            tags=tags,
            abortable=abortable,
            monitored=run_monitoring_thread,
            git_info=git_info,
            hostname=hostname,
            entrypoint=entrypoint,
            notebook_id=notebook_id,
            checkpoint_id=checkpoint_id
        )

        self._backend.upload_source_code(experiment, source_target_pairs)

        # pylint: disable=protected-access
        experiment._start(
            abort_callback=abort_callback,
            logger=logger,
            upload_stdout=upload_stdout,
            upload_stderr=upload_stderr,
            send_hardware_metrics=send_hardware_metrics,
            run_monitoring_thread=run_monitoring_thread,
            handle_uncaught_exceptions=handle_uncaught_exceptions
        )

        self._push_new_experiment(experiment)

        click.echo(self._get_experiment_link(experiment))

        return experiment

    def _get_experiment_link(self, experiment):
        return "{base_url}/{namespace}/{project}/e/{exp_id}".format(
            base_url=self._backend.display_address,
            namespace=self.namespace,
            project=self.name,
            exp_id=experiment.id
        )

    def create_notebook(self):
        """Create a new notebook object and return corresponding :class:`~neptune.notebook.Notebook` instance.

        Returns:
            :class:`~neptune.notebook.Notebook` object.

        Examples:

            .. code:: python3

                # Instantiate a session and fetch a project
                project = neptune.init()

                # Create a notebook in Neptune
                notebook = project.create_notebook()
        """
        return self._backend.create_notebook(self)

    def get_notebook(self, notebook_id):
        """Get a :class:`~neptune.notebook.Notebook` object with given ``notebook_id``.

        Returns:
            :class:`~neptune.notebook.Notebook` object.

        Examples:

            .. code:: python3

                # Instantiate a session and fetch a project
                project = neptune.init()

                # Get a notebook object
                notebook = project.get_notebook('d1c1b494-0620-4e54-93d5-29f4e848a51a')
        """
        return self._backend.get_notebook(project=self, notebook_id=notebook_id)

    @property
    def full_id(self):
        """Project qualified name as :obj:`str`, for example `john/sandbox`.
        """
        return '{}/{}'.format(self.namespace, self.name)

    def __str__(self):
        return 'Project({})'.format(self.full_id)

    def __repr__(self):
        return str(self)

    def __eq__(self, o):
        return self.__dict__ == o.__dict__

    def __ne__(self, o):
        return not self.__eq__(o)

    def _fetch_leaderboard(self, id, state, owner, tag, min_running_time):
        return self._backend.get_leaderboard_entries(
            project=self, ids=as_list(id), states=as_list(state),
            owners=as_list(owner), tags=as_list(tag),
            min_running_time=min_running_time)

    @staticmethod
    def _sort_leaderboard_columns(column_names):
        user_defined_weights = {
            'channel': 1,
            'parameter': 2,
            'property': 3
        }

        system_properties_weights = {
            'id': 0,
            'name': 1,
            'created': 2,
            'finished': 3,
            'owner': 4,
            'worker_type': 5,
            'environment': 6,
        }

        def key(c):
            """A sorting key for a column name.

            Sorts by the system properties first, then channels, parameters, user-defined properties.

            Within each group columns are sorted alphabetically, except for system properties,
            where order is custom.
            """
            parts = c.split('_', 1)
            if parts[0] in user_defined_weights.keys():
                name = parts[1]
                weight = user_defined_weights.get(parts[0], 99)
                system_property_weight = None
            else:
                name = c
                weight = 0
                system_property_weight = system_properties_weights.get(name, 99)

            return weight, system_property_weight, name

        return sorted(column_names, key=key)

    def _get_current_experiment(self):
        with self.__lock:
            if self._experiments_stack:
                return self._experiments_stack[-1]
            else:
                raise NeptuneNoExperimentContextException()

    def _push_new_experiment(self, new_experiment):
        with self.__lock:
            self._experiments_stack.append(new_experiment)
            return new_experiment

    def _remove_stopped_experiment(self, experiment):
        with self.__lock:
            if self._experiments_stack:
                self._experiments_stack = [exp for exp in self._experiments_stack if exp != experiment]

    def _shutdown_hook(self):
        if self._experiments_stack:
            # stopping experiment removes it from list, co we copy it
            copied_experiment_list = [exp for exp in self._experiments_stack]
            for exp in copied_experiment_list:
                exp.stop()



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import base64
import json
import os
from typing import Optional, Dict

from dataclasses import dataclass

from neptune.new import envs
from neptune.new import ANONYMOUS, ANONYMOUS_API_TOKEN
from neptune.new.exceptions import NeptuneInvalidApiTokenException, NeptuneMissingApiTokenException


@dataclass(frozen=True)
class Credentials:
    api_token: str
    token_origin_address: str
    api_url_opt: str

    @classmethod
    def from_token(cls, api_token: Optional[str] = None) -> 'Credentials':
        if api_token is None:
            api_token = os.getenv(envs.API_TOKEN_ENV_NAME)

        if api_token == ANONYMOUS:
            api_token = ANONYMOUS_API_TOKEN

        if api_token is None:
            raise NeptuneMissingApiTokenException()

        token_dict = Credentials._api_token_to_dict(api_token)
        # TODO: Consider renaming 'api_address' (breaking backward compatibility)
        if 'api_address' not in token_dict:
            raise NeptuneInvalidApiTokenException()
        token_origin_address = token_dict['api_address']
        api_url = token_dict['api_url'] if 'api_url' in token_dict else None

        return Credentials(
            api_token=api_token,
            token_origin_address=token_origin_address,
            api_url_opt=api_url
        )

    @staticmethod
    def _api_token_to_dict(api_token: str) -> Dict[str, str]:
        try:
            return json.loads(base64.b64decode(api_token.encode()).decode("utf-8"))
        except Exception:
            raise NeptuneInvalidApiTokenException()



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import abc
from dataclasses import dataclass
from datetime import datetime
from typing import List, TypeVar, Generic, Optional, Set, Tuple
from typing import TYPE_CHECKING

from neptune.new.exceptions import InternalClientError

if TYPE_CHECKING:
    from neptune.new.internal.operation_visitor import OperationVisitor

Ret = TypeVar('Ret')
T = TypeVar('T')


def all_subclasses(cls):
    return set(cls.__subclasses__()).union(
        [s for c in cls.__subclasses__() for s in all_subclasses(c)])


@dataclass
class Operation:

    path: List[str]

    @abc.abstractmethod
    def accept(self, visitor: 'OperationVisitor[Ret]') -> Ret:
        pass

    # pylint: disable=unused-argument
    def to_dict(self) -> dict:
        return {
            "type": self.__class__.__name__,
            "path": self.path
        }

    @staticmethod
    def from_dict(data: dict) -> 'Operation':
        if "type" not in data:
            raise ValueError("Malformed operation {} - type is missing".format(data))
        sub_classes = {cls.__name__: cls for cls in all_subclasses(Operation)}
        if not data["type"] in sub_classes:
            raise ValueError("Malformed operation {} - unknown type {}".format(data, data["type"]))
        return sub_classes[data["type"]].from_dict(data)


@dataclass
class AssignFloat(Operation):

    value: float

    def accept(self, visitor: 'OperationVisitor[Ret]') -> Ret:
        return visitor.visit_assign_float(self)

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret["value"] = self.value
        return ret

    @staticmethod
    def from_dict(data: dict) -> 'AssignFloat':
        return AssignFloat(data["path"], data["value"])


@dataclass
class AssignInt(Operation):

    value: int

    def accept(self, visitor: 'OperationVisitor[Ret]') -> Ret:
        return visitor.visit_assign_int(self)

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret["value"] = self.value
        return ret

    @staticmethod
    def from_dict(data: dict) -> 'AssignInt':
        return AssignInt(data["path"], data["value"])


@dataclass
class AssignBool(Operation):

    value: bool

    def accept(self, visitor: 'OperationVisitor[Ret]') -> Ret:
        return visitor.visit_assign_bool(self)

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret["value"] = self.value
        return ret

    @staticmethod
    def from_dict(data: dict) -> 'AssignBool':
        return AssignBool(data["path"], data["value"])


@dataclass
class AssignString(Operation):

    value: str

    def accept(self, visitor: 'OperationVisitor[Ret]') -> Ret:
        return visitor.visit_assign_string(self)

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret["value"] = self.value
        return ret

    @staticmethod
    def from_dict(data: dict) -> 'AssignString':
        return AssignString(data["path"], data["value"])


@dataclass
class AssignDatetime(Operation):

    value: datetime

    def accept(self, visitor: 'OperationVisitor[Ret]') -> Ret:
        return visitor.visit_assign_datetime(self)

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret["value"] = int(1000 * self.value.timestamp())
        return ret

    @staticmethod
    def from_dict(data: dict) -> 'AssignDatetime':
        return AssignDatetime(data["path"], datetime.fromtimestamp(data["value"] / 1000))


@dataclass
class AssignArtifact(Operation):

    hash: str

    def accept(self, visitor: 'OperationVisitor[Ret]') -> Ret:
        return visitor.visit_assign_artifact(self)

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret["hash"] = self.hash
        return ret

    @staticmethod
    def from_dict(data: dict) -> 'AssignArtifact':
        return AssignArtifact(data["path"], str(data["hash"]))


@dataclass
class UploadFile(Operation):

    ext: str
    file_path: str

    def accept(self, visitor: 'OperationVisitor[Ret]') -> Ret:
        return visitor.visit_upload_file(self)

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret["ext"] = self.ext
        ret["file_path"] = self.file_path
        return ret

    @staticmethod
    def from_dict(data: dict) -> 'UploadFile':
        return UploadFile(data["path"], data["ext"], data["file_path"])


@dataclass
class UploadFileContent(Operation):

    ext: str
    file_content: str

    def accept(self, visitor: 'OperationVisitor[Ret]') -> Ret:
        return visitor.visit_upload_file_content(self)

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret["ext"] = self.ext
        ret["file_content"] = self.file_content
        return ret

    @staticmethod
    def from_dict(data: dict) -> 'UploadFileContent':
        return UploadFileContent(data["path"], data["ext"], data["file_content"])


@dataclass
class UploadFileSet(Operation):

    file_globs: List[str]
    reset: bool

    def accept(self, visitor: 'OperationVisitor[Ret]') -> Ret:
        return visitor.visit_upload_file_set(self)

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret["file_globs"] = self.file_globs
        ret["reset"] = str(self.reset)
        return ret

    @staticmethod
    def from_dict(data: dict) -> 'UploadFileSet':
        return UploadFileSet(data["path"], data["file_globs"], data["reset"] != str(False))


@dataclass
class LogSeriesValue(Generic[T]):

    value: T
    step: Optional[float]
    ts: float

    def to_dict(self, value_serializer=lambda x: x) -> dict:
        return {
            "value": value_serializer(self.value),
            "step": self.step,
            "ts": self.ts
        }

    @staticmethod
    def from_dict(data: dict, value_deserializer=lambda x: x) -> 'LogSeriesValue[T]':
        return LogSeriesValue[T](value_deserializer(data["value"]), data.get("step", None), data["ts"])


@dataclass
class LogFloats(Operation):

    ValueType = LogSeriesValue[float]

    values: List[ValueType]

    def accept(self, visitor: 'OperationVisitor[Ret]') -> Ret:
        return visitor.visit_log_floats(self)

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret["values"] = [value.to_dict() for value in self.values]
        return ret

    @staticmethod
    def from_dict(data: dict) -> 'LogFloats':
        return LogFloats(
            data["path"],
            [LogFloats.ValueType.from_dict(value) for value in data["values"]]
        )


@dataclass
class LogStrings(Operation):

    ValueType = LogSeriesValue[str]

    values: List[ValueType]

    def accept(self, visitor: 'OperationVisitor[Ret]') -> Ret:
        return visitor.visit_log_strings(self)

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret["values"] = [value.to_dict() for value in self.values]
        return ret

    @staticmethod
    def from_dict(data: dict) -> 'LogStrings':
        return LogStrings(
            data["path"],
            [LogStrings.ValueType.from_dict(value) for value in data["values"]]
        )


@dataclass
class ImageValue:
    data: Optional[str]
    name: Optional[str]
    description: Optional[str]

    @staticmethod
    def serializer(obj: 'ImageValue'):
        return dict(data=obj.data, name=obj.name, description=obj.description)

    @staticmethod
    def deserializer(obj) -> 'ImageValue':
        if obj is None:
            return ImageValue(None, None, None)
        if isinstance(obj, str):
            return ImageValue(data=obj, name=None, description=None)
        if isinstance(obj, dict):
            return ImageValue(data=obj["data"], name=obj["name"], description=obj["description"])
        else:
            raise InternalClientError("Run data on disk is malformed or was saved by newer version of Neptune Library")


@dataclass
class LogImages(Operation):

    ValueType = LogSeriesValue[ImageValue]

    values: List[ValueType]

    def accept(self, visitor: 'OperationVisitor[Ret]') -> Ret:
        return visitor.visit_log_images(self)

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret["values"] = [value.to_dict(ImageValue.serializer) for value in self.values]
        return ret

    @staticmethod
    def from_dict(data: dict) -> 'LogImages':
        return LogImages(
            data["path"],
            [LogImages.ValueType.from_dict(value, ImageValue.deserializer) for value in data["values"]]
        )


@dataclass
class ClearFloatLog(Operation):

    def accept(self, visitor: 'OperationVisitor[Ret]') -> Ret:
        return visitor.visit_clear_float_log(self)

    @staticmethod
    def from_dict(data: dict) -> 'ClearFloatLog':
        return ClearFloatLog(data["path"])


@dataclass
class ClearStringLog(Operation):

    def accept(self, visitor: 'OperationVisitor[Ret]') -> Ret:
        return visitor.visit_clear_string_log(self)

    @staticmethod
    def from_dict(data: dict) -> 'ClearStringLog':
        return ClearStringLog(data["path"])


@dataclass
class ClearImageLog(Operation):

    def accept(self, visitor: 'OperationVisitor[Ret]') -> Ret:
        return visitor.visit_clear_image_log(self)

    @staticmethod
    def from_dict(data: dict) -> 'ClearImageLog':
        return ClearImageLog(data["path"])


@dataclass
class ConfigFloatSeries(Operation):

    min: Optional[float]
    max: Optional[float]
    unit: Optional[str]

    def accept(self, visitor: 'OperationVisitor[Ret]') -> Ret:
        return visitor.visit_config_float_series(self)

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret["min"] = self.min
        ret["max"] = self.max
        ret["unit"] = self.unit
        return ret

    @staticmethod
    def from_dict(data: dict) -> 'ConfigFloatSeries':
        return ConfigFloatSeries(data["path"], data["min"], data["max"], data["unit"])


@dataclass
class AddStrings(Operation):

    values: Set[str]

    def accept(self, visitor: 'OperationVisitor[Ret]') -> Ret:
        return visitor.visit_add_strings(self)

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret["values"] = list(self.values)
        return ret

    @staticmethod
    def from_dict(data: dict) -> 'AddStrings':
        return AddStrings(data["path"], set(data["values"]))


@dataclass
class RemoveStrings(Operation):

    values: Set[str]

    def accept(self, visitor: 'OperationVisitor[Ret]') -> Ret:
        return visitor.visit_remove_strings(self)

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret["values"] = list(self.values)
        return ret

    @staticmethod
    def from_dict(data: dict) -> 'RemoveStrings':
        return RemoveStrings(data["path"], set(data["values"]))


@dataclass
class ClearStringSet(Operation):

    def accept(self, visitor: 'OperationVisitor[Ret]') -> Ret:
        return visitor.visit_clear_string_set(self)

    @staticmethod
    def from_dict(data: dict) -> 'ClearStringSet':
        return ClearStringSet(data["path"])


@dataclass
class DeleteFiles(Operation):

    file_paths: Set[str]

    def accept(self, visitor: 'OperationVisitor[Ret]') -> Ret:
        return visitor.visit_delete_files(self)

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret["file_paths"] = list(self.file_paths)
        return ret

    @staticmethod
    def from_dict(data: dict) -> 'DeleteFiles':
        return DeleteFiles(data["path"], set(data["file_paths"]))


@dataclass
class DeleteAttribute(Operation):

    def accept(self, visitor: 'OperationVisitor[Ret]') -> Ret:
        return visitor.visit_delete_attribute(self)

    @staticmethod
    def from_dict(data: dict) -> 'DeleteAttribute':
        return DeleteAttribute(data["path"])


@dataclass
class TrackFilesToArtifact(Operation):
    project_id: str
    entries: List[Tuple[str, Optional[str]]]

    def accept(self, visitor: 'OperationVisitor[Ret]') -> Ret:
        return visitor.visit_track_files_to_artifact(self)

    def to_dict(self) -> dict:
        ret = super().to_dict()
        ret["entries"] = self.entries
        ret["project_id"] = self.project_id
        return ret

    @staticmethod
    def from_dict(data: dict) -> 'TrackFilesToArtifact':
        return TrackFilesToArtifact(
            path=data["path"],
            project_id=data["project_id"],
            entries=list(map(tuple, data["entries"]))
        )


@dataclass
class ClearArtifact(Operation):

    def accept(self, visitor: 'OperationVisitor[Ret]') -> Ret:
        return visitor.visit_clear_artifact(self)

    @staticmethod
    def from_dict(data: dict) -> 'ClearArtifact':
        return ClearArtifact(data["path"])



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import logging
import os
import sys
from typing import Optional, List

from neptune.new import Run
from neptune.new.attributes import constants as attr_consts
from neptune.new.internal.utils import get_absolute_paths, get_common_root
from neptune.internal.storage.storage_utils import normalize_file_name
from neptune.utils import is_ipython

_logger = logging.getLogger(__name__)


def upload_source_code(source_files: Optional[List[str]], run: Run) -> None:
    if not is_ipython() and os.path.isfile(sys.argv[0]):
        if source_files is None:
            entrypoint = os.path.basename(sys.argv[0])
            source_files = sys.argv[0]
        elif not source_files:
            entrypoint = os.path.basename(sys.argv[0])
        else:
            common_root = get_common_root(get_absolute_paths(source_files))
            if common_root is not None:
                entrypoint = normalize_file_name(os.path.relpath(os.path.abspath(sys.argv[0]), common_root))
            else:
                entrypoint = normalize_file_name(os.path.abspath(sys.argv[0]))
        run[attr_consts.SOURCE_CODE_ENTRYPOINT_ATTRIBUTE_PATH] = entrypoint

    if source_files is not None:
        run[attr_consts.SOURCE_CODE_FILES_ATTRIBUTE_PATH].upload_files(source_files)



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# pylint: disable=protected-access
import numpy
from mock import patch, MagicMock, call
from neptune.new.exceptions import OperationNotSupported
from neptune.new.internal.utils import base64_encode
from neptune.new.internal.operation import ImageValue, LogImages, ClearImageLog
from neptune.new.types import File
from neptune.new.attributes.series.file_series import FileSeries
from tests.neptune.new.attributes.test_attribute_base import TestAttributeBase
from tests.neptune.new.helpers import create_file


@patch("time.time", new=TestAttributeBase._now)
class TestFileSeries(TestAttributeBase):

    def test_assign_type_error(self):
        values = [[5.], ["text"], 55, "string", None]
        for value in values:
            with self.assertRaises(Exception):
                FileSeries(MagicMock(), MagicMock()).assign(value)

    def test_log_type_error(self):
        values = [[5.], [[]], 55, None]
        for value in values:
            with self.assertRaises(TypeError):
                FileSeries(MagicMock(), MagicMock()).log(value)

    def test_log_content(self):
        # given
        wait = self._random_wait()
        path = self._random_path()
        op_processor = MagicMock()
        exp = self._create_run(processor=op_processor)
        attr = FileSeries(exp, path)

        file = File.as_image(numpy.random.rand(10, 10) * 255)

        # when
        attr.log(file, step=3, timestamp=self._now(), wait=wait, name="nazwa", description="opis")

        # then
        op_processor.enqueue_operation.assert_called_once_with(
            LogImages(path, [LogImages.ValueType(
                ImageValue(base64_encode(file.content), "nazwa", "opis"),
                3,
                self._now())]), wait
        )

    def test_assign_content(self):
        # given
        wait = self._random_wait()
        path = self._random_path()
        op_processor = MagicMock()
        exp = self._create_run(processor=op_processor)
        attr = FileSeries(exp, path)

        file = File.as_image(numpy.random.rand(10, 10) * 255)

        # when
        attr.assign([file], wait=wait)

        # then
        op_processor.enqueue_operation.assert_has_calls([
            call(ClearImageLog(path), False),
            call(LogImages(path, [LogImages.ValueType(
                ImageValue(base64_encode(file.content), None, None),
                None,
                self._now())]), wait),
        ])

    def test_log_path(self):
        # given
        wait = self._random_wait()
        path = self._random_path()
        op_processor = MagicMock()
        exp = self._create_run(processor=op_processor)
        attr = FileSeries(exp, path)

        file = File.as_image(numpy.random.rand(10, 10) * 255)
        with create_file(file.content, binary_mode=True) as tmp_filename:
            saved_file = File(tmp_filename)

            # when
            attr.log(saved_file, step=3, timestamp=self._now(), wait=wait, description="something")

            # then
            op_processor.enqueue_operation.assert_called_once_with(
                LogImages(path, [LogImages.ValueType(
                    ImageValue(base64_encode(file.content), None, "something"),
                    3,
                    self._now())]), wait
            )

    def test_log_raise_not_image(self):
        # given
        path = self._random_path()
        op_processor = MagicMock()
        exp = self._create_run(processor=op_processor)
        attr = FileSeries(exp, path)

        file = File.from_content("some text")
        with create_file(file.content, binary_mode=True) as tmp_filename:
            saved_file = File(tmp_filename)

            # when
            with self.assertRaises(OperationNotSupported):
                attr.log(file)
            with self.assertRaises(OperationNotSupported):
                attr.log(saved_file)

    def test_assign_raise_not_image(self):
        # given
        path = self._random_path()
        op_processor = MagicMock()
        exp = self._create_run(processor=op_processor)
        attr = FileSeries(exp, path)

        file = File.from_content("some text")
        with create_file(file.content, binary_mode=True) as tmp_filename:
            saved_file = File(tmp_filename)

            # when
            with self.assertRaises(OperationNotSupported):
                attr.assign([file])
            with self.assertRaises(OperationNotSupported):
                attr.assign([saved_file])



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import typing
import hashlib
import datetime
from pathlib import Path

from neptune.new.internal.artifacts.types import ArtifactFileData, ArtifactMetadataSerializer
from neptune.new.internal.artifacts.local_file_hash_storage import LocalFileHashStorage
from neptune.new.internal.artifacts.utils import sha1


class FileHasher:
    ENCODING = "UTF-8"
    HASH_ELEMENT_DIVISOR = b"#"
    META_ELEMENT_DIVISOR = b"|"
    SERVER_INT_BYTES = 4
    SERVER_LONG_BYTES = 8
    SERVER_BYTE_ORDER = 'big'
    HASH_LENGTH = 64  # sha-256

    @classmethod
    def get_local_file_hash(cls, file_path: typing.Union[str, Path]) -> str:
        local_storage = LocalFileHashStorage()

        absolute = Path(file_path).resolve()
        modification_date = datetime.datetime.fromtimestamp(absolute.stat().st_mtime).strftime('%Y%m%d_%H%M%S%f')

        stored_file_hash = local_storage.fetch_one(absolute)

        if stored_file_hash:
            if stored_file_hash.modification_date >= modification_date:
                return stored_file_hash.file_hash
            else:
                computed_hash = sha1(absolute)
                local_storage.update(absolute, computed_hash, modification_date)

                return computed_hash
        else:
            computed_hash = sha1(absolute)
            local_storage.insert(absolute, computed_hash, modification_date)

            return computed_hash

    @classmethod
    def _number_to_bytes(cls, int_value: int, bytes_cnt):
        return int_value.to_bytes(bytes_cnt, cls.SERVER_BYTE_ORDER)

    @classmethod
    def get_artifact_hash(cls, artifact_files: typing.Iterable[ArtifactFileData]) -> str:
        artifact_hash = hashlib.sha256()

        for artifact_file in sorted(artifact_files, key=lambda file: file.file_path):
            artifact_hash.update(cls.HASH_ELEMENT_DIVISOR)
            artifact_hash.update(cls._number_to_bytes(len(artifact_file.file_path), cls.SERVER_INT_BYTES))
            artifact_hash.update(artifact_file.file_path.encode(cls.ENCODING))
            artifact_hash.update(cls.HASH_ELEMENT_DIVISOR)
            artifact_hash.update(artifact_file.file_hash.encode(cls.ENCODING))
            artifact_hash.update(cls.HASH_ELEMENT_DIVISOR)
            if artifact_file.size is not None:
                artifact_hash.update(cls._number_to_bytes(artifact_file.size, cls.SERVER_LONG_BYTES))
            artifact_hash.update(cls.HASH_ELEMENT_DIVISOR)
            artifact_hash.update(cls._number_to_bytes(len(artifact_file.type), cls.SERVER_INT_BYTES))
            artifact_hash.update(artifact_file.type.encode(cls.ENCODING))
            artifact_hash.update(cls.HASH_ELEMENT_DIVISOR)
            for metadata_key_value in ArtifactMetadataSerializer.serialize(artifact_file.metadata):
                metadata_name, metadata_value = metadata_key_value.get('key'), metadata_key_value.get('value')
                artifact_hash.update(cls.META_ELEMENT_DIVISOR)
                artifact_hash.update(cls._number_to_bytes(len(metadata_name), cls.SERVER_INT_BYTES))
                artifact_hash.update(metadata_name.encode(cls.ENCODING))
                artifact_hash.update(cls.META_ELEMENT_DIVISOR)
                artifact_hash.update(cls._number_to_bytes(len(metadata_value), cls.SERVER_INT_BYTES))
                artifact_hash.update(metadata_value.encode(cls.ENCODING))

        return str(artifact_hash.hexdigest())



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# pylint: disable=protected-access
from datetime import datetime

from mock import MagicMock

from neptune.new.internal.operation import AssignDatetime
from neptune.new.attributes.atoms.datetime import Datetime, DatetimeVal

from tests.neptune.new.attributes.test_attribute_base import TestAttributeBase


class TestDatetime(TestAttributeBase):

    def test_assign(self):
        now = datetime.now()
        value_and_expected = [
            (now, now.replace(microsecond=1000*int(now.microsecond/1000))),
            (DatetimeVal(now), now.replace(microsecond=1000*int(now.microsecond/1000)))
        ]

        for value, expected in value_and_expected:
            processor = MagicMock()
            exp, path, wait = self._create_run(processor), self._random_path(), self._random_wait()
            var = Datetime(exp, path)
            var.assign(value, wait=wait)
            processor.enqueue_operation.assert_called_once_with(AssignDatetime(path, expected), wait)

    def test_assign_type_error(self):
        values = [55, None]
        for value in values:
            with self.assertRaises(TypeError):
                Datetime(MagicMock(), MagicMock()).assign(value)

    def test_get(self):
        exp, path = self._create_run(), self._random_path()
        var = Datetime(exp, path)
        now = datetime.now()
        now = now.replace(microsecond=int(now.microsecond/1000)*1000)
        var.assign(now)
        self.assertEqual(now, var.fetch())



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# pylint: disable=protected-access
import os
import unittest
from io import StringIO, BytesIO

from mock import MagicMock

from neptune.new.attributes.atoms.file import File, FileVal
from neptune.new.attributes.file_set import FileSet, FileSetVal
from neptune.new.internal.operation import UploadFile, UploadFileSet, UploadFileContent
from neptune.new.internal.utils import base64_encode
from neptune.utils import IS_WINDOWS
from tests.neptune.new.attributes.test_attribute_base import TestAttributeBase


class TestFile(TestAttributeBase):

    @unittest.skipIf(IS_WINDOWS, "Windows behaves strangely")
    def test_assign(self):
        a_text = "Some text stream"
        a_binary = b"Some binary stream"
        value_and_operation_factory = [
            (FileVal("other/../other/file.txt"),
             lambda _: UploadFile(_, "txt", os.getcwd() + "/other/file.txt")),
            (FileVal.from_stream(StringIO(a_text)),
             lambda _: UploadFileContent(_, "txt", base64_encode(a_text.encode('utf-8')))),
            (FileVal.from_stream(BytesIO(a_binary)),
             lambda _: UploadFileContent(_, "bin", base64_encode(a_binary))),
        ]

        for value, operation_factory in value_and_operation_factory:
            processor = MagicMock()
            exp, path, wait = self._create_run(processor), self._random_path(), self._random_wait()
            var = File(exp, path)
            var.assign(value, wait=wait)
            processor.enqueue_operation.assert_called_once_with(operation_factory(path), wait)

    def test_assign_type_error(self):
        values = [55, None, []]
        for value in values:
            with self.assertRaises(TypeError):
                File(MagicMock(), MagicMock()).assign(value)

    @unittest.skipIf(IS_WINDOWS, "Windows behaves strangely")
    def test_save(self):
        value_and_expected = [
            ("some/path", os.getcwd() + "/some/path")
        ]

        for value, expected in value_and_expected:
            processor = MagicMock()
            exp, path, wait = self._create_run(processor), self._random_path(), self._random_wait()
            var = File(exp, path)
            var.upload(value, wait=wait)
            processor.enqueue_operation.assert_called_once_with(UploadFile(path=path, ext="", file_path=expected),
                                                                wait)

    @unittest.skipIf(IS_WINDOWS, "Windows behaves strangely")
    def test_save_files(self):
        value_and_expected = [
            ("some/path/*", [os.getcwd() + "/some/path/*"])
        ]

        for value, expected in value_and_expected:
            processor = MagicMock()
            exp, path, wait = self._create_run(processor), self._random_path(), self._random_wait()
            var = FileSet(exp, path)
            var.upload_files(value, wait=wait)
            processor.enqueue_operation.assert_called_once_with(UploadFileSet(path, expected, False), wait)

    def test_save_type_error(self):
        values = [55, None, [], FileVal]
        for value in values:
            with self.assertRaises(TypeError):
                File(MagicMock(), MagicMock()).upload(value)

    def test_save__files_type_error(self):
        values = [55, None, [55], FileSetVal]
        for value in values:
            with self.assertRaises(TypeError):
                FileSet(MagicMock(), MagicMock()).upload_files(value)

    def test_fetch_extension(self):
        value_and_expected_ext = [
            (FileVal("some/file.txt"), "txt"),
            (FileVal("some/file"), ""),
            (FileVal.from_content("Some text stream"), "txt"),
            (FileVal.from_content(b"Some binary stream"), "bin"),
            (FileVal.from_content(b"Some binary stream", extension="png"), "png")
        ]

        for value, expected_ext in value_and_expected_ext:
            exp, path, wait = self._create_run(), self._random_path(), self._random_wait()
            var = File(exp, path)
            var.assign(value, wait=wait)
            self.assertEqual(expected_ext, var.fetch_extension())
